from typing import Iterator, Mapping, Sequence, Type

from apischema.conversions.conversions import DefaultConversion
from apischema.conversions.visitor import DeserializationVisitor
from apischema.objects import ObjectField
from apischema.objects.visitor import DeserializationObjectVisitor
from apischema.types import AnyType
from apischema.utils import get_origin_or_type
from apischema.visitor import Unsupported


class InitFlattenedAliasVisitor(
    DeserializationObjectVisitor[Iterator[str]], DeserializationVisitor[Iterator[str]]
):
    def mapping(
        self, cls: Type[Mapping], key_type: AnyType, value_type: AnyType
    ) -> Iterator[str]:
        yield from ()

    def object(self, tp: AnyType, fields: Sequence[ObjectField]) -> Iterator[str]:
        for field in fields:
            if field.flattened:
                yield from get_deserialization_flattened_aliases(
                    get_origin_or_type(tp), field, self.default_conversion
                )
            elif not field.is_aggregate:
                yield field.alias

    def _visited_union(self, results: Sequence[Iterator[str]]) -> Iterator[str]:
        if len(results) != 1:
            raise NotImplementedError
        return results[0]


def get_deserialization_flattened_aliases(
    cls: Type, field: ObjectField, default_conversion: DefaultConversion
) -> Iterator[str]:
    assert field.flattened
    try:
        yield from InitFlattenedAliasVisitor(default_conversion).visit_with_conv(
            field.type, field.deserialization
        )
    except (NotImplementedError, Unsupported):
        raise TypeError(
            f"Flattened field {cls.__name__}.{field.name} must have an object type"
        ) from None



import collections.abc
import operator
from contextlib import suppress
from dataclasses import dataclass
from enum import Enum
from functools import lru_cache
from typing import (
    Any,
    Callable,
    Collection,
    Mapping,
    Optional,
    Sequence,
    Tuple,
    Type,
    TypeVar,
    Union,
    cast,
    overload,
)

from apischema.aliases import Aliaser
from apischema.cache import cache
from apischema.conversions.conversions import AnyConversion, DefaultConversion
from apischema.conversions.utils import Converter
from apischema.conversions.visitor import (
    Serialization,
    SerializationVisitor,
    sub_conversion,
)
from apischema.fields import FIELDS_SET_ATTR, support_fields_set
from apischema.objects import AliasedStr, ObjectField
from apischema.objects.visitor import SerializationObjectVisitor
from apischema.ordering import sort_by_order
from apischema.recursion import RecursiveConversionsVisitor
from apischema.serialization.serialized_methods import get_serialized_methods
from apischema.types import AnyType, NoneType, Undefined, UndefinedType
from apischema.typing import is_new_type, is_type, is_type_var, is_typed_dict, is_union
from apischema.utils import (
    Lazy,
    as_predicate,
    deprecate_kwargs,
    get_args2,
    get_origin_or_type,
    get_origin_or_type2,
    identity,
    is_union_of,
    opt_or,
)
from apischema.visitor import Unsupported

SerializationMethod = Callable[[Any], Any]
SerializationMethodFactory = Callable[[AnyType], SerializationMethod]


T = TypeVar("T")


def instance_checker(tp: AnyType) -> Tuple[Callable[[Any, Any], bool], Any]:
    origin = get_origin_or_type2(tp)
    if origin is NoneType:
        return operator.is_, None
    elif is_typed_dict(origin):
        return isinstance, collections.abc.Mapping
    elif is_type(origin):
        return isinstance, origin
    elif is_new_type(origin):
        return instance_checker(origin.__supertype__)
    elif is_type_var(origin) or origin is Any:
        return (lambda data, _: True), ...
    elif is_union(origin):
        checks = list(map(instance_checker, get_args2(tp)))
        return (lambda data, _: any(check(data, arg) for check, arg in checks)), ...
    else:
        raise TypeError(f"{tp} is not supported in union serialization")


def identity_as_none(method: SerializationMethod) -> Optional[SerializationMethod]:
    return method if method is not identity else None


@dataclass(frozen=True)
class PassThroughOptions:
    any: bool = False
    collections: bool = False
    enums: bool = False
    types: Union[Collection[AnyType], Callable[[AnyType], bool]] = ()

    def __post_init__(self):
        object.__setattr__(self, "types", as_predicate(self.types))


class SerializationMethodVisitor(
    RecursiveConversionsVisitor[Serialization, SerializationMethod],
    SerializationVisitor[SerializationMethod],
    SerializationObjectVisitor[SerializationMethod],
):
    use_cache: bool = True

    def __init__(
        self,
        additional_properties: bool,
        aliaser: Aliaser,
        check_type: bool,
        default_conversion: DefaultConversion,
        exclude_defaults: bool,
        exclude_none: bool,
        exclude_unset: bool,
        fall_back_on_any: bool,
        pass_through_options: PassThroughOptions,
    ):
        super().__init__(default_conversion)
        self.additional_properties = additional_properties
        self.aliaser = aliaser
        self.check_type = check_type
        self.exclude_defaults = exclude_defaults
        self.exclude_none = exclude_none
        self.exclude_unset = exclude_unset
        self.fall_back_on_any = fall_back_on_any
        self.pass_through_options = pass_through_options
        self.pass_through_type = as_predicate(self.pass_through_options.types)

    @property
    def _factory(self) -> SerializationMethodFactory:
        return serialization_method_factory(
            self.additional_properties,
            self.aliaser,
            self.check_type,
            self._conversion,
            self.default_conversion,
            self.exclude_defaults,
            self.exclude_none,
            self.exclude_unset,
            self.fall_back_on_any,
            self.pass_through_options,
        )

    def visit_not_recursive(self, tp: AnyType):
        return self._factory(tp) if self.use_cache else super().visit_not_recursive(tp)

    def _recursive_result(self, lazy: Lazy[SerializationMethod]) -> SerializationMethod:
        rec_method = None

        def method(obj: Any) -> Any:
            nonlocal rec_method
            if rec_method is None:
                rec_method = lazy()
            return rec_method(obj)

        return method

    def any(self) -> SerializationMethod:
        if self.pass_through_options.any:
            return identity
        factory = self._factory

        def method(obj: Any) -> Any:
            return factory(obj.__class__)(obj)

        return method

    def _any_fallback(self, tp: AnyType) -> SerializationMethod:
        fallback, serialize_any = self.fall_back_on_any, self.any()

        def method(obj: Any) -> Any:
            if fallback:
                return serialize_any(obj)
            else:
                raise TypeError(f"Expected {tp}, found {obj.__class__}")

        return method

    def _wrap(self, cls: type, method: SerializationMethod) -> SerializationMethod:
        if not self.check_type:
            return method
        fallback = self._any_fallback(cls)
        cls_to_check = Mapping if is_typed_dict(cls) else cls

        def wrapper(obj: Any) -> Any:
            if isinstance(obj, cls_to_check):
                try:
                    return method(obj)
                except Exception:
                    pass
            return fallback(obj)

        return wrapper

    def collection(
        self, cls: Type[Collection], value_type: AnyType
    ) -> SerializationMethod:
        serialize_value = self.visit(value_type)

        method: SerializationMethod
        if serialize_value is not identity:

            def method(obj: Any) -> Any:
                # using map is faster than comprehension
                return list(map(serialize_value, obj))

        elif issubclass(cls, (list, tuple)) or (
            self.pass_through_options.collections
            and not issubclass(cls, collections.abc.Set)
        ):
            method = identity
        else:
            method = list

        return self._wrap(cls, method)

    def enum(self, cls: Type[Enum]) -> SerializationMethod:
        if self.pass_through_options.enums or issubclass(cls, (int, str)):
            return identity
        elif all(
            method is identity
            for method in map(self.visit, {elt.value.__class__ for elt in cls})
        ):
            method: SerializationMethod = operator.attrgetter("value")
        else:
            any_method = self.any()

            def method(obj: Any) -> Any:
                return any_method(obj.value)

        return self._wrap(cls, method)

    def literal(self, values: Sequence[Any]) -> SerializationMethod:
        if self.pass_through_options.enums or all(
            isinstance(v, (int, str)) for v in values
        ):
            return identity
        else:
            return self.any()

    def mapping(
        self, cls: Type[Mapping], key_type: AnyType, value_type: AnyType
    ) -> SerializationMethod:
        serialize_key, serialize_value = self.visit(key_type), self.visit(value_type)
        method: SerializationMethod
        if serialize_key is not identity or serialize_value is not identity:

            def method(obj: Any) -> Any:
                return {
                    serialize_key(key): serialize_value(value)
                    for key, value in obj.items()
                }

        elif self.pass_through_options.collections or issubclass(cls, dict):
            method = identity
        else:
            method = dict

        return self._wrap(cls, method)

    def object(self, tp: AnyType, fields: Sequence[ObjectField]) -> SerializationMethod:
        cls = get_origin_or_type(tp)
        typed_dict = is_typed_dict(cls)
        getter: Callable[[str], Callable[[Any], Any]] = (
            operator.itemgetter if typed_dict else operator.attrgetter
        )
        serialization_fields = [
            (
                field.name,
                self.aliaser(field.alias) if not field.is_aggregate else None,
                getter(field.name),
                field.required,
                field.skip.serialization_if,
                is_union_of(field.type, UndefinedType) or default is Undefined,
                (is_union_of(field.type, NoneType) and self.exclude_none)
                or field.none_as_undefined
                or (default is None and self.exclude_defaults),
                (field.skip.serialization_default or self.exclude_defaults)
                and default not in (None, Undefined),
                default,
                identity_as_none(self.visit_with_conv(field.type, field.serialization)),
                field.ordering,
            )
            for field in fields
            for default in [... if field.required else field.get_default()]
        ] + [
            (
                serialized.func.__name__,
                self.aliaser(serialized.alias),
                serialized.func,
                True,
                None,
                is_union_of(ret_type, UndefinedType),
                is_union_of(ret_type, NoneType) and self.exclude_none,
                False,
                ...,
                self.visit_with_conv(ret_type, serialized.conversion),
                serialized.ordering,
            )
            for serialized, types in get_serialized_methods(tp)
            for ret_type in [types["return"]]
        ]
        serialization_fields = sort_by_order(  # type: ignore
            cls, serialization_fields, lambda f: f[0], lambda f: f[-1]
        )
        field_names = {f.name for f in fields}
        any_method = self.any()
        exclude_unset = self.exclude_unset and support_fields_set(cls)
        additional_properties = self.additional_properties and typed_dict

        def method(obj: Any) -> Any:
            result = {}
            for (
                name,
                alias,
                get_field,
                required,
                skip_if,
                undefined,
                skip_none,
                skip_default,
                default,
                serialize_field,
                _,
            ) in serialization_fields:
                if (not exclude_unset or name in getattr(obj, FIELDS_SET_ATTR)) and (
                    not typed_dict or required or name in obj
                ):
                    field_value = get_field(obj)
                    if not (
                        (skip_if and skip_if(field_value))
                        or (undefined and field_value is Undefined)
                        or (skip_none and field_value is None)
                        or (skip_default and field_value == default)
                    ):
                        if serialize_field:
                            field_value = serialize_field(field_value)
                        if alias:
                            result[alias] = field_value
                        else:
                            result.update(field_value)
            if additional_properties:
                assert isinstance(obj, Mapping)
                for key, value in obj.items():
                    if key not in field_names and isinstance(key, str):
                        result[key] = any_method(value)
            return result

        return self._wrap(cls, method)

    def primitive(self, cls: Type) -> SerializationMethod:
        return self._wrap(cls, identity)

    def subprimitive(self, cls: Type, superclass: Type) -> SerializationMethod:
        if cls is AliasedStr:
            return self.aliaser
        else:
            return super().subprimitive(cls, superclass)

    def tuple(self, types: Sequence[AnyType]) -> SerializationMethod:
        elt_serializers = list(enumerate(map(self.visit, types)))
        if all(method is identity for _, method in elt_serializers):
            return identity

        def method(obj: Any) -> Any:
            return [serialize_elt(obj[i]) for i, serialize_elt in elt_serializers]

        if self.check_type:
            nb_elts = len(elt_serializers)
            wrapped = method
            fall_back_on_any, as_list = self.fall_back_on_any, self._factory(list)

            def method(obj: Any) -> Any:
                if len(obj) == nb_elts:
                    return wrapped(obj)
                elif fall_back_on_any:
                    return as_list(obj)
                else:
                    raise TypeError(f"Expected {nb_elts}-tuple, found {len(obj)}-tuple")

        return self._wrap(tuple, method)

    def union(self, alternatives: Sequence[AnyType]) -> SerializationMethod:
        methods = []
        for tp in alternatives:
            with suppress(Unsupported):
                methods.append((self.visit(tp), *instance_checker(tp)))
        # No need to catch the case with all methods being identity,
        # because passthrough
        if not methods:
            raise Unsupported(Union[tuple(alternatives)])  # type: ignore
        elif len(methods) == 1:
            return methods[0][0]
        elif all(method is identity for method, _, _ in methods):
            return identity
        elif len(methods) == 2 and NoneType in alternatives:
            serialize_alt = next(meth for meth, _, arg in methods if arg is not None)

            def method(obj: Any) -> Any:
                return serialize_alt(obj) if obj is not None else None

        else:
            fallback = self._any_fallback(Union[alternatives])

            def method(obj: Any) -> Any:
                for serialize_alt, check, arg in methods:
                    if check(obj, arg):
                        try:
                            return serialize_alt(obj)
                        except Exception:
                            pass
                return fallback(obj)

        return method

    def unsupported(self, tp: AnyType) -> SerializationMethod:
        try:
            return super().unsupported(tp)
        except Unsupported:
            if self.fall_back_on_any and is_type(tp):
                if issubclass(tp, Mapping):
                    return self.visit(Mapping[Any, Any])
                elif issubclass(tp, Collection):
                    return self.visit(Collection[Any])
            raise

    def _visit_conversion(
        self,
        tp: AnyType,
        conversion: Serialization,
        dynamic: bool,
        next_conversion: Optional[AnyConversion],
    ) -> SerializationMethod:
        serialize_conv = self.visit_with_conv(
            conversion.target, sub_conversion(conversion, next_conversion)
        )
        converter = cast(Converter, conversion.converter)
        if converter is identity:
            method = serialize_conv
        elif serialize_conv is identity:
            method = converter
        else:

            def method(obj: Any) -> Any:
                return serialize_conv(converter(obj))

        return self._wrap(get_origin_or_type(tp), method)

    def visit_conversion(
        self,
        tp: AnyType,
        conversion: Optional[Serialization],
        dynamic: bool,
        next_conversion: Optional[AnyConversion] = None,
    ) -> SerializationMethod:
        if not dynamic and self.pass_through_type(tp):
            return identity
        else:
            return super().visit_conversion(tp, conversion, dynamic, next_conversion)


@cache
def serialization_method_factory(
    additional_properties: bool,
    aliaser: Aliaser,
    check_type: bool,
    conversion: Optional[AnyConversion],
    default_conversion: DefaultConversion,
    exclude_defaults: bool,
    exclude_none: bool,
    exclude_unset: bool,
    fall_back_on_any: bool,
    pass_through: PassThroughOptions,
) -> SerializationMethodFactory:
    @lru_cache()
    def factory(tp: AnyType) -> SerializationMethod:
        return SerializationMethodVisitor(
            additional_properties,
            aliaser,
            check_type,
            default_conversion,
            exclude_defaults,
            exclude_none,
            exclude_unset,
            fall_back_on_any,
            pass_through,
        ).visit_with_conv(tp, conversion)

    return factory


def serialization_method(
    type: AnyType,
    *,
    additional_properties: bool = None,
    aliaser: Aliaser = None,
    check_type: bool = None,
    conversion: AnyConversion = None,
    default_conversion: DefaultConversion = None,
    exclude_defaults: bool = None,
    exclude_none: bool = None,
    exclude_unset: bool = None,
    fall_back_on_any: bool = None,
    pass_through: PassThroughOptions = None,
) -> SerializationMethod:
    from apischema import settings

    return serialization_method_factory(
        opt_or(additional_properties, settings.additional_properties),
        opt_or(aliaser, settings.aliaser),
        opt_or(check_type, settings.serialization.check_type),
        conversion,
        opt_or(default_conversion, settings.serialization.default_conversion),
        opt_or(exclude_defaults, settings.serialization.exclude_defaults),
        opt_or(exclude_none, settings.serialization.exclude_none),
        opt_or(exclude_unset, settings.serialization.exclude_unset),
        opt_or(fall_back_on_any, settings.serialization.fall_back_on_any),
        opt_or(pass_through, settings.serialization.pass_through),
    )(type)


NO_OBJ = object()


@overload
def serialize(
    type: AnyType,
    obj: Any,
    *,
    additional_properties: bool = None,
    aliaser: Aliaser = None,
    check_type: bool = None,
    conversion: AnyConversion = None,
    default_conversion: DefaultConversion = None,
    exclude_defaults: bool = None,
    exclude_none: bool = None,
    exclude_unset: bool = None,
    fall_back_on_any: bool = None,
    pass_through: PassThroughOptions = None,
) -> Any:
    ...


@overload
def serialize(
    obj: Any,
    *,
    additional_properties: bool = None,
    aliaser: Aliaser = None,
    check_type: bool = None,
    conversion: AnyConversion = None,
    default_conversion: DefaultConversion = None,
    exclude_defaults: bool = None,
    exclude_none: bool = None,
    exclude_unset: bool = None,
    fall_back_on_any: bool = True,
    pass_through: PassThroughOptions = None,
) -> Any:
    ...


@deprecate_kwargs({"conversions": "conversion"})  # type: ignore
def serialize(
    type: AnyType = Any,
    obj: Any = NO_OBJ,
    *,
    additional_properties: bool = None,
    aliaser: Aliaser = None,
    check_type: bool = None,
    conversion: AnyConversion = None,
    default_conversion: DefaultConversion = None,
    exclude_defaults: bool = None,
    exclude_none: bool = None,
    exclude_unset: bool = None,
    fall_back_on_any: bool = None,
    pass_through: PassThroughOptions = None,
) -> Any:
    # Handle overloaded signature without type
    if obj is NO_OBJ:
        type, obj = Any, type
        if fall_back_on_any is None:
            fall_back_on_any = True
    return serialization_method(
        type,
        additional_properties=additional_properties,
        aliaser=aliaser,
        check_type=check_type,
        conversion=conversion,
        default_conversion=default_conversion,
        exclude_defaults=exclude_defaults,
        exclude_none=exclude_none,
        exclude_unset=exclude_unset,
        fall_back_on_any=fall_back_on_any,
        pass_through=pass_through,
    )(obj)


def serialization_default(
    *,
    additional_properties: bool = None,
    aliaser: Aliaser = None,
    default_conversion: DefaultConversion = None,
    exclude_defaults: bool = None,
    exclude_none: bool = None,
    exclude_unset: bool = None,
) -> SerializationMethod:
    from apischema import settings

    factory = serialization_method_factory(
        opt_or(additional_properties, settings.additional_properties),
        opt_or(aliaser, settings.aliaser),
        False,
        None,
        opt_or(default_conversion, settings.serialization.default_conversion),
        opt_or(exclude_defaults, settings.serialization.exclude_defaults),
        opt_or(exclude_none, settings.serialization.exclude_none),
        opt_or(exclude_unset, settings.serialization.exclude_unset),
        False,
        PassThroughOptions(any=True),
    )

    def method(obj: Any) -> Any:
        return factory(obj.__class__)(obj)

    return method



import sys
from collections import defaultdict
from enum import Enum
from functools import partial
from types import new_class
from typing import (
    Callable,
    List,
    MutableMapping,
    Optional,
    TYPE_CHECKING,
    Type,
    TypeVar,
    Union,
    overload,
)

from apischema.cache import CacheAwareDict
from apischema.conversions import LazyConversion
from apischema.conversions.conversions import (
    AnyConversion,
    ConvOrFunc,
    Conversion,
    resolve_conversion,
)
from apischema.conversions.utils import Converter, is_convertible
from apischema.methods import MethodOrProperty, MethodWrapper, is_method, method_class
from apischema.type_names import type_name
from apischema.types import AnyType
from apischema.typing import is_type_var
from apischema.utils import get_args2, get_origin_or_type, stop_signature_abuse

if TYPE_CHECKING:
    pass


_deserializers: MutableMapping[AnyType, List[ConvOrFunc]] = CacheAwareDict(
    defaultdict(list)
)
_serializers: MutableMapping[AnyType, ConvOrFunc] = CacheAwareDict({})
Deserializer = TypeVar(
    "Deserializer", bound=Union[Callable, Conversion, staticmethod, type]
)
Serializer = TypeVar("Serializer", bound=Union[Callable, Conversion, property, type])

default_deserialization: Callable[[type], Optional[AnyConversion]]
# defaultdict.get is not hashable in 3.7
if sys.version_info < (3, 8):

    def default_deserialization(tp):
        return _deserializers.get(tp)


else:
    default_deserialization = _deserializers.get  # type: ignore


def default_serialization(tp: Type) -> Optional[AnyConversion]:
    for sub_cls in getattr(tp, "__mro__", [tp]):
        if sub_cls in _serializers:
            conversion = _serializers[sub_cls]
            if (
                sub_cls == tp
                or not isinstance(conversion, (Conversion, LazyConversion))
                or conversion.inherited in (None, True)
            ):
                return conversion
    else:
        return None


def check_converter_type(tp: AnyType) -> AnyType:
    origin = get_origin_or_type(tp)
    if not is_convertible(tp):
        raise TypeError(f"{origin} is not convertible")
    if not all(map(is_type_var, get_args2(tp))):
        raise TypeError("Generic conversion doesn't support specialization")
    return origin


def _add_deserializer(conversion: ConvOrFunc, target: AnyType):
    target = check_converter_type(target)
    if conversion not in _deserializers[target]:
        _deserializers[target].append(conversion)


class DeserializerDescriptor(MethodWrapper[staticmethod]):
    def __set_name__(self, owner, name):
        super().__set_name__(owner, name)
        method = self._method.__get__(None, object)
        resolved = resolve_conversion(method, {owner.__name__: owner})
        _add_deserializer(method, resolved.target)


@overload
def deserializer(deserializer: Deserializer) -> Deserializer:
    ...


@overload
def deserializer(
    *, lazy: Callable[[], Union[Converter, Conversion]], target: Type
) -> None:
    ...


def deserializer(
    deserializer: Deserializer = None,
    *,
    lazy: Callable[[], Union[Converter, Conversion]] = None,
    target: Type = None,
):
    if deserializer is not None:
        if isinstance(deserializer, staticmethod):
            return DeserializerDescriptor(deserializer)
        elif isinstance(deserializer, LazyConversion):
            stop_signature_abuse()
        else:
            resolved = resolve_conversion(deserializer)  # type: ignore
            _add_deserializer(deserializer, resolved.target)  # type: ignore
            return deserializer
    elif lazy is not None and target is not None:
        _add_deserializer(LazyConversion(lazy), target)
    else:
        stop_signature_abuse()


def _add_serializer(conversion: ConvOrFunc, source: AnyType):
    source = check_converter_type(source)
    _serializers[source] = conversion


class SerializerDescriptor(MethodWrapper[MethodOrProperty]):
    def __set_name__(self, owner, name):
        super().__set_name__(owner, name)
        _add_serializer(self._method, source=owner)


@overload
def serializer(serializer: Serializer) -> Serializer:
    ...


@overload
def serializer(
    *, lazy: Callable[[], Union[Converter, Conversion]], source: Type
) -> Callable[[Serializer], Serializer]:
    ...


def serializer(
    serializer: Serializer = None,
    *,
    lazy: Callable[[], Union[Converter, Conversion]] = None,
    source: Type = None,
):
    if serializer is not None:
        if is_method(serializer) and method_class(serializer) is None:  # type: ignore
            return SerializerDescriptor(serializer)  # type: ignore
        elif isinstance(serializer, LazyConversion):
            stop_signature_abuse()
        else:
            resolved = resolve_conversion(serializer)
            _add_serializer(serializer, resolved.source)
            return serializer
    elif lazy is not None and source is not None:
        _add_serializer(LazyConversion(lazy), source)
    else:
        stop_signature_abuse()


def reset_deserializers(cls: Type):
    _deserializers.pop(cls, ...)


def reset_serializer(cls: Type):
    _serializers.pop(cls, ...)


Cls = TypeVar("Cls", bound=type)


def as_str(cls: Cls) -> Cls:
    deserializer(Conversion(cls, source=str))
    serializer(Conversion(str, source=cls))
    return cls


EnumCls = TypeVar("EnumCls", bound=Type[Enum])


def as_names(cls: EnumCls, aliaser: Callable[[str], str] = lambda s: s) -> EnumCls:
    # Enum requires to call namespace __setitem__
    def exec_body(namespace: dict):
        for elt in cls:  # type: ignore
            namespace[elt.name] = aliaser(elt.name)

    if not issubclass(cls, Enum):
        raise TypeError("as_names must be called with Enum subclass")
    name_cls = type_name(None)(
        new_class(cls.__name__, (str, Enum), exec_body=exec_body)
    )
    deserializer(Conversion(partial(getattr, cls), source=name_cls, target=cls))

    def get_name(obj):
        return getattr(name_cls, obj.name)

    serializer(Conversion(get_name, source=cls, target=name_cls))
    return cls



from dataclasses import dataclass
from typing import Generic, Optional, Sequence, Type, TypeVar

from apischema.type_names import get_type_name, type_name
from apischema.types import NoneType
from apischema.typing import generic_mro, get_args, get_origin
from apischema.utils import get_args2, is_union_of, wrap_generic_init_subclass

Cursor_ = TypeVar("Cursor_")
Node_ = TypeVar("Node_")


def get_node_name(tp):
    if is_union_of(tp, NoneType) and len(get_args2(tp)):
        tp = next(arg for arg in get_args2(tp) if arg is not NoneType)
    ref = get_type_name(tp).graphql
    if ref is None:
        raise TypeError(
            f"Node {tp} must have a ref registered to be used with connection"
        )
    return ref


def edge_name(tp: Type["Edge"], *args) -> str:
    for base in generic_mro(tp[tuple(args)] if args else tp):  # type: ignore
        if get_origin(base) == Edge:
            return f"{get_node_name(get_args(base)[0])}Edge"
    raise NotImplementedError


@type_name(graphql=edge_name)
@dataclass
class Edge(Generic[Node_, Cursor_]):
    node: Node_
    cursor: Cursor_

    @wrap_generic_init_subclass
    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        type_name(graphql=edge_name)(cls)


@type_name(graphql=lambda *_: "PageInfo")
@dataclass
class PageInfo(Generic[Cursor_]):
    has_previous_page: bool = False
    has_next_page: bool = False
    start_cursor: Optional[Cursor_] = None
    end_cursor: Optional[Cursor_] = None

    @staticmethod
    def from_edges(
        edges: Sequence[Optional[Edge[Node_, Cursor_]]],
        has_previous_page: bool = False,
        has_next_page: bool = False,
    ) -> "PageInfo":
        start_cursor, end_cursor = None, None
        if edges is not None:
            if edges[0] is not None:
                start_cursor = edges[0].cursor
            if edges[-1] is not None:
                end_cursor = edges[-1].cursor
        return PageInfo(has_previous_page, has_next_page, start_cursor, end_cursor)


def connection_name(tp: Type["Connection"], *args) -> str:
    for base in generic_mro(tp[tuple(args)] if args else tp):  # type: ignore
        if get_origin(base) == Connection:
            return f"{get_node_name(get_args(base)[0])}Connection"
    raise NotImplementedError


Edge_ = TypeVar("Edge_", bound=Edge)


@type_name(graphql=connection_name)
@dataclass
class Connection(Generic[Node_, Cursor_, Edge_]):
    edges: Optional[Sequence[Optional[Edge_]]]
    page_info: PageInfo[Cursor_]

    @wrap_generic_init_subclass
    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        type_name(graphql=connection_name)(cls)



from collections import defaultdict
from dataclasses import dataclass
from typing import (
    AbstractSet,
    Any,
    Collection,
    Dict,
    List,
    Mapping,
    MutableMapping,
    Set,
    Tuple,
    overload,
)

from apischema.cache import CacheAwareDict
from apischema.objects.fields import check_field_or_name, get_field_name

_dependent_requireds: MutableMapping[
    type, List[Tuple[Any, Collection[Any]]]
] = CacheAwareDict(defaultdict(list))

DependentRequired = Mapping[str, AbstractSet[str]]


def get_dependent_required(cls: type) -> DependentRequired:
    result: Dict[str, Set[str]] = defaultdict(set)
    for sub_cls in cls.__mro__:
        for field, required in _dependent_requireds[sub_cls]:
            result[get_field_name(field)].update(map(get_field_name, required))
    return result


@dataclass
class DependentRequiredDescriptor:
    fields: Mapping[Any, Collection[Any]]
    groups: Collection[Collection[Any]]

    def __set_name__(self, owner, name):
        setattr(owner, name, None)
        dependent_required(self.fields, *self.groups, owner=owner)


@overload
def dependent_required(
    fields: Mapping[Any, Collection[Any]], *groups: Collection[Any], owner: type = None
):
    ...


@overload
def dependent_required(*groups: Collection[Any], owner: type = None):
    ...


def dependent_required(*groups: Collection[Any], owner: type = None):  # type: ignore
    if not groups:
        return
    fields: Mapping[Any, Collection[Any]] = {}
    if isinstance(groups[0], Mapping):
        fields, *groups = groups  # type: ignore
    if owner is None:
        return DependentRequiredDescriptor(fields, groups)
    else:

        dep_req = _dependent_requireds[owner]
        for field, required in fields.items():
            dep_req.append((field, required))
            check_field_or_name(field)
            for req in required:
                check_field_or_name(req)
        for group in map(list, groups):
            for i, field in enumerate(group):
                check_field_or_name(field)
                dep_req.append((field, [group[:i], group[i:]]))



from dataclasses import dataclass
from typing import Any, Callable, Dict, Optional, Sequence, Union

from apischema.utils import to_camel_case

Deprecated = Union[bool, str]
try:
    from apischema.typing import Literal

    ContentEncoding = Literal["7bit", "8bit", "binary", "quoted-printable", "base64"]
except ImportError:
    ContentEncoding = str  # type: ignore


@dataclass(frozen=True)
class Annotations:
    title: Optional[str] = None
    description: Optional[str] = None
    default: Optional[Callable[[], Any]] = None
    examples: Optional[Sequence[Any]] = None
    format: Optional[str] = None
    deprecated: Optional[Deprecated] = None
    media_type: Optional[str] = None
    encoding: Optional[ContentEncoding] = None

    def merge_into(self, base_schema: Dict[str, Any]):
        if self.deprecated:
            base_schema["deprecated"] = bool(self.deprecated)
        for k in ("title", "description", "examples", "format"):
            if getattr(self, k) is not None:
                base_schema[k] = getattr(self, k)
        for k in ("media_type", "encoding"):
            if getattr(self, k) is not None:
                base_schema[to_camel_case("content_" + k)] = getattr(self, k)
        if self.default is not None:
            base_schema["default"] = self.default()



__all__ = [
    "PassThroughOptions",
    "Undefined",
    "UndefinedType",
    "Unsupported",
    "ValidationError",
    "alias",
    "dependent_required",
    "deserialization_method",
    "deserialize",
    "deserializer",
    "identity",
    "order",
    "properties",
    "schema",
    "schema_ref",
    "serialization_default",
    "serialization_method",
    "serialize",
    "serialized",
    "serializer",
    "settings",
    "type_name",
    "validator",
]

import warnings

from . import (  # noqa: F401
    cache,
    conversions,
    dataclasses,
    fields,
    json_schema,
    metadata,
    objects,
    skip,
    tagged_unions,
    validation,
)
from .aliases import alias
from .conversions import deserializer, serializer
from .dependencies import dependent_required
from .deserialization import deserialization_method, deserialize
from .metadata import properties
from .ordering import order
from .schemas import schema
from .serialization import (
    PassThroughOptions,
    serialization_default,
    serialization_method,
    serialize,
)
from .serialization.serialized_methods import serialized
from .settings import settings
from .type_names import schema_ref, type_name
from .types import Undefined, UndefinedType
from .utils import identity
from .validation import ValidationError, validator
from .visitor import Unsupported

try:
    from . import graphql  # noqa: F401

    __all__.append("graphql")
except ImportError:
    pass


def __getattr__(name):
    if name == "graphql":
        raise AttributeError(
            "GraphQL feature requires graphql-core library\n"
            "Run `pip install apischema[graphql]` to install it"
        )
    if name == "skip":
        warnings.warn("apischema.skip module is deprecated")
    raise AttributeError(f"module {__name__!r} has no attribute {name!r}")


def register_default_conversions():
    """Handle standard library + internal types"""
    from . import std_types  # noqa: F401

    deserializer(ValidationError.from_errors)
    serializer(ValidationError.errors)


register_default_conversions()
del register_default_conversions



from contextlib import suppress
from dataclasses import dataclass
from enum import Enum
from itertools import chain
from typing import (
    AbstractSet,
    Any,
    Callable,
    ClassVar,
    Collection,
    Dict,
    List,
    Mapping,
    Optional,
    Pattern,
    Sequence,
    Tuple,
    Type,
    TypeVar,
    Union,
)

from apischema.aliases import Aliaser
from apischema.conversions import converters
from apischema.conversions.conversions import AnyConversion, DefaultConversion
from apischema.conversions.visitor import (
    Conv,
    ConversionsVisitor,
    Deserialization,
    DeserializationVisitor,
    Serialization,
    SerializationVisitor,
)
from apischema.dependencies import get_dependent_required
from apischema.json_schema.conversions_resolver import WithConversionsResolver
from apischema.json_schema.patterns import infer_pattern
from apischema.json_schema.refs import (
    DeserializationRefsExtractor,
    Refs,
    RefsExtractor as RefsExtractor_,
    SerializationRefsExtractor,
)
from apischema.json_schema.types import JsonSchema, JsonType, json_schema
from apischema.json_schema.versions import JsonSchemaVersion, RefFactory
from apischema.metadata.keys import SCHEMA_METADATA
from apischema.objects import AliasedStr, ObjectField
from apischema.objects.visitor import (
    DeserializationObjectVisitor,
    ObjectVisitor,
    SerializationObjectVisitor,
)
from apischema.ordering import Ordering, sort_by_order
from apischema.schemas import Schema, get_schema as _get_schema, merge_schema
from apischema.serialization import serialize
from apischema.serialization.serialized_methods import (
    SerializedMethod,
    get_serialized_methods,
)
from apischema.type_names import TypeNameFactory, get_type_name
from apischema.types import AnyType, OrderedDict, UndefinedType
from apischema.typing import get_args, is_typed_dict
from apischema.utils import (
    context_setter,
    deprecate_kwargs,
    get_origin_or_type,
    is_union_of,
    literal_values,
)


def get_schema(tp: AnyType) -> Optional[Schema]:
    from apischema import settings

    return merge_schema(settings.base_schema.type(tp), _get_schema(tp))


def get_field_schema(tp: AnyType, field: ObjectField) -> Optional[Schema]:
    from apischema import settings

    assert not field.is_aggregate
    return merge_schema(
        settings.base_schema.field(tp, field.name, field.alias), field.schema
    )


def get_method_schema(tp: AnyType, method: SerializedMethod) -> Optional[Schema]:
    from apischema import settings

    return merge_schema(
        settings.base_schema.method(tp, method.func, method.alias), method.schema
    )


def full_schema(base_schema: JsonSchema, schema: Optional[Schema]) -> JsonSchema:
    if schema is not None:
        base_schema = JsonSchema(base_schema)
        schema.merge_into(base_schema)
    return base_schema


Method = TypeVar("Method", bound=Callable)


@dataclass(frozen=True)
class Property:
    alias: AliasedStr
    name: str
    ordering: Optional[Ordering]
    required: bool
    schema: JsonSchema


class SchemaBuilder(
    ConversionsVisitor[Conv, JsonSchema],
    ObjectVisitor[JsonSchema],
    WithConversionsResolver,
):
    def __init__(
        self,
        additional_properties: bool,
        default_conversion: DefaultConversion,
        ignore_first_ref: bool,
        ref_factory: RefFactory,
        refs: Collection[str],
    ):
        super().__init__(default_conversion)
        self.additional_properties = additional_properties
        self._ignore_first_ref = ignore_first_ref
        self.ref_factory = ref_factory
        self.refs = refs

    def ref_schema(self, ref: Optional[str]) -> Optional[JsonSchema]:
        if ref not in self.refs:
            return None
        elif self._ignore_first_ref:
            self._ignore_first_ref = False
            return None
        else:
            assert isinstance(ref, str)
            return JsonSchema({"$ref": self.ref_factory(ref)})

    def annotated(self, tp: AnyType, annotations: Sequence[Any]) -> JsonSchema:
        schema = None
        for annotation in reversed(annotations):
            if isinstance(annotation, TypeNameFactory):
                ref = annotation.to_type_name(tp).json_schema
                ref_schema = self.ref_schema(ref)
                if ref_schema is not None:
                    return full_schema(ref_schema, schema)
            if isinstance(annotation, Mapping):
                schema = merge_schema(annotation.get(SCHEMA_METADATA), schema)
        return full_schema(super().annotated(tp, annotations), schema)

    def any(self) -> JsonSchema:
        return JsonSchema()

    def collection(self, cls: Type[Collection], value_type: AnyType) -> JsonSchema:
        return json_schema(
            type=JsonType.ARRAY,
            items=self.visit(value_type),
            uniqueItems=issubclass(cls, AbstractSet),
        )

    def enum(self, cls: Type[Enum]) -> JsonSchema:
        if len(cls) == 0:
            raise TypeError("Empty enum")
        return self.literal(list(cls))

    def literal(self, values: Sequence[Any]) -> JsonSchema:
        if not values:
            raise TypeError("Empty Literal")
        types = {JsonType.from_type(type(v)) for v in literal_values(values)}
        # Mypy issue
        type_: Any = types.pop() if len(types) == 1 else types
        if len(values) == 1:
            return json_schema(type=type_, const=values[0])
        else:
            return json_schema(type=type_, enum=values)

    def mapping(
        self, cls: Type[Mapping], key_type: AnyType, value_type: AnyType
    ) -> JsonSchema:
        with context_setter(self):
            self._ignore_first_ref = True
            key = self.visit(key_type)
        if key["type"] != JsonType.STRING:
            raise ValueError("Mapping types must string-convertible key")
        value = self.visit(value_type)
        if "pattern" in key:
            return json_schema(
                type=JsonType.OBJECT, patternProperties={key["pattern"]: value}
            )
        else:
            return json_schema(type=JsonType.OBJECT, additionalProperties=value)

    def visit_field(
        self, tp: AnyType, field: ObjectField, required: bool = True
    ) -> JsonSchema:
        assert not field.is_aggregate
        result = full_schema(
            self.visit_with_conv(field.type, self._field_conversion(field)),
            get_field_schema(tp, field) if tp is not None else field.schema,
        )
        if not required and "default" not in result:
            result = JsonSchema(result)
            with suppress(Exception):
                result["default"] = serialize(
                    field.type,
                    field.get_default(),
                    fall_back_on_any=False,
                    check_type=True,
                    conversion=field.serialization,
                )
        return result

    def _object_schema(self, cls: type, field: ObjectField) -> JsonSchema:
        assert field.is_aggregate
        with context_setter(self):
            self._ignore_first_ref = True
            object_schema = full_schema(
                self.visit_with_conv(field.type, self._field_conversion(field)),
                field.schema,
            )
        if object_schema.get("type") not in {JsonType.OBJECT, "object"}:
            field_type = "Flattened" if field.flattened else "Properties"
            raise TypeError(
                f"{field_type} field {cls.__name__}.{field.name}"
                f" must have an object type"
            )
        return object_schema

    def _properties_schema(
        self, object_schema: JsonSchema, pattern: Optional[Pattern] = None
    ):
        if "patternProperties" in object_schema:
            if pattern is not None:
                for p in (pattern, pattern.pattern):
                    if p in object_schema["patternProperties"]:
                        return object_schema["patternProperties"][p]
            elif (
                len(object_schema["patternProperties"]) == 1
                and "additionalProperties" not in object_schema
            ):
                return next(iter(object_schema["patternProperties"].values()))
        if isinstance(object_schema.get("additionalProperties"), Mapping):
            return object_schema["additionalProperties"]
        return JsonSchema()

    def properties(
        self, tp: AnyType, fields: Sequence[ObjectField]
    ) -> Sequence[Property]:
        raise NotImplementedError

    def object(self, tp: AnyType, fields: Sequence[ObjectField]) -> JsonSchema:
        cls = get_origin_or_type(tp)
        properties = sort_by_order(
            cls, self.properties(tp, fields), lambda p: p.name, lambda p: p.ordering
        )
        flattened_schemas: List[JsonSchema] = []
        pattern_properties = {}
        additional_properties: Union[bool, JsonSchema] = self.additional_properties
        for field in fields:
            if field.flattened:
                self._object_schema(cls, field)  # check the field is an object
                flattened_schemas.append(
                    full_schema(
                        self.visit_with_conv(field.type, self._field_conversion(field)),
                        field.schema,
                    )
                )
            elif field.pattern_properties is not None:
                if field.pattern_properties is ...:
                    pattern = infer_pattern(field.type, self.default_conversion)
                else:
                    assert isinstance(field.pattern_properties, Pattern)
                    pattern = field.pattern_properties
                pattern_properties[pattern] = self._properties_schema(
                    self._object_schema(cls, field), pattern
                )
            elif field.additional_properties:
                additional_properties = self._properties_schema(
                    self._object_schema(cls, field)
                )
        alias_by_names = {f.name: f.alias for f in fields}.__getitem__
        dependent_required = get_dependent_required(cls)
        result = json_schema(
            type=JsonType.OBJECT,
            properties={p.alias: p.schema for p in properties},
            required=[p.alias for p in properties if p.required],
            additionalProperties=additional_properties,
            patternProperties=pattern_properties,
            dependentRequired=OrderedDict(
                (alias_by_names(f), sorted(map(alias_by_names, dependent_required[f])))
                for f in sorted(dependent_required, key=alias_by_names)
            ),
        )
        if flattened_schemas:
            result = json_schema(
                type=JsonType.OBJECT,
                allOf=[result, *flattened_schemas],
                unevaluatedProperties=False,
            )
        return result

    def primitive(self, cls: Type) -> JsonSchema:
        return JsonSchema(type=JsonType.from_type(cls))

    def tuple(self, types: Sequence[AnyType]) -> JsonSchema:
        return json_schema(
            type=JsonType.ARRAY,
            prefixItems=[self.visit(cls) for cls in types],
            items=False,
            minItems=len(types),
            maxItems=len(types),
        )

    def _visited_union(self, results: Sequence[JsonSchema]) -> JsonSchema:
        if len(results) == 1:
            return results[0]
        elif any(alt == {} for alt in results):
            return JsonSchema()
        elif all(alt.keys() == {"type"} for alt in results):
            types: Any = chain.from_iterable(
                [res["type"]]
                if isinstance(res["type"], (str, JsonType))
                else res["type"]
                for res in results
            )
            return json_schema(type=list(types))
        elif (
            len(results) == 2
            and all("type" in res for res in results)
            and {"type": "null"} in results
        ):
            for result in results:
                if result != {"type": "null"}:
                    types = result["type"]
                    if isinstance(types, (str, JsonType)):
                        types = [types]
                    if "null" not in types:
                        result = JsonSchema({**result, "type": [*types, "null"]})
                    return result
            else:
                raise NotImplementedError
        else:
            return json_schema(anyOf=results)

    def visit_conversion(
        self,
        tp: AnyType,
        conversion: Optional[Conv],
        dynamic: bool,
        next_conversion: Optional[AnyConversion] = None,
    ) -> JsonSchema:
        schema = None
        if not dynamic:
            for ref_tp in self.resolve_conversion(tp):
                ref_schema = self.ref_schema(get_type_name(ref_tp).json_schema)
                if ref_schema is not None:
                    return ref_schema
            if get_args(tp):
                schema = merge_schema(schema, get_schema(get_origin_or_type(tp)))
            schema = merge_schema(schema, get_schema(tp))
        result = super().visit_conversion(tp, conversion, dynamic, next_conversion)
        return full_schema(result, schema)

    RefsExtractor: ClassVar[Type[RefsExtractor_]]


class DeserializationSchemaBuilder(
    SchemaBuilder[Deserialization],
    DeserializationVisitor[JsonSchema],
    DeserializationObjectVisitor[JsonSchema],
):
    RefsExtractor = DeserializationRefsExtractor

    def properties(
        self, tp: AnyType, fields: Sequence[ObjectField]
    ) -> Sequence[Property]:
        return [
            Property(
                AliasedStr(field.alias),
                field.name,
                field.ordering,
                field.required,
                self.visit_field(tp, field, field.required),
            )
            for field in fields
            if not field.is_aggregate
        ]


class SerializationSchemaBuilder(
    SchemaBuilder[Serialization],
    SerializationVisitor[JsonSchema],
    SerializationObjectVisitor[JsonSchema],
):
    RefsExtractor = SerializationRefsExtractor

    @staticmethod
    def _field_required(field: ObjectField):
        from apischema import settings

        return not field.skippable(
            settings.serialization.exclude_defaults, settings.serialization.exclude_none
        )

    def properties(
        self, tp: AnyType, fields: Sequence[ObjectField]
    ) -> Sequence[Property]:
        from apischema import settings

        return [
            Property(
                AliasedStr(field.alias),
                field.name,
                field.ordering,
                required,
                self.visit_field(tp, field, required),
            )
            for field in fields
            if not field.is_aggregate
            for required in [
                field.required
                if is_typed_dict(get_origin_or_type(tp))
                else not field.skippable(
                    settings.serialization.exclude_defaults,
                    settings.serialization.exclude_none,
                )
            ]
        ] + [
            Property(
                AliasedStr(serialized.alias),
                serialized.func.__name__,
                serialized.ordering,
                not is_union_of(types["return"], UndefinedType),
                full_schema(
                    self.visit_with_conv(types["return"], serialized.conversion),
                    get_method_schema(tp, serialized),
                ),
            )
            for serialized, types in get_serialized_methods(tp)
        ]


TypesWithConversion = Collection[Union[AnyType, Tuple[AnyType, AnyConversion]]]


def _default_version(
    version: Optional[JsonSchemaVersion],
    ref_factory: Optional[RefFactory],
    all_refs: Optional[bool],
) -> Tuple[JsonSchemaVersion, RefFactory, bool]:
    from apischema import settings

    if version is None:
        version = settings.json_schema_version
    if ref_factory is None:
        ref_factory = version.ref_factory
    if all_refs is None:
        all_refs = version.all_refs
    return version, ref_factory, all_refs


def _extract_refs(
    types: TypesWithConversion,
    default_conversion: DefaultConversion,
    builder: Type[SchemaBuilder],
    all_refs: bool,
) -> Mapping[str, AnyType]:
    refs: Refs = {}
    for tp in types:
        conversion = None
        if isinstance(tp, tuple):
            tp, conversion = tp
        builder.RefsExtractor(default_conversion, refs).visit_with_conv(tp, conversion)
    filtr = (lambda count: True) if all_refs else (lambda count: count > 1)
    return {ref: tp for ref, (tp, count) in refs.items() if filtr(count)}


def _refs_schema(
    builder: Type[SchemaBuilder],
    default_conversion: DefaultConversion,
    refs: Mapping[str, AnyType],
    ref_factory: RefFactory,
    additional_properties: bool,
) -> Mapping[str, JsonSchema]:
    return {
        ref: builder(
            additional_properties, default_conversion, True, ref_factory, refs
        ).visit(tp)
        for ref, tp in refs.items()
    }


def _schema(
    builder: Type[SchemaBuilder],
    tp: AnyType,
    schema: Optional[Schema],
    conversion: Optional[AnyConversion],
    default_conversion: DefaultConversion,
    version: Optional[JsonSchemaVersion],
    aliaser: Optional[Aliaser],
    ref_factory: Optional[RefFactory],
    all_refs: Optional[bool],
    with_schema: bool,
    additional_properties: Optional[bool],
) -> Mapping[str, Any]:
    from apischema import settings

    add_defs = ref_factory is None
    if aliaser is None:
        aliaser = settings.aliaser
    if additional_properties is None:
        additional_properties = settings.additional_properties
    version, ref_factory, all_refs = _default_version(version, ref_factory, all_refs)
    refs = _extract_refs([(tp, conversion)], default_conversion, builder, all_refs)
    json_schema = builder(
        additional_properties, default_conversion, False, ref_factory, refs
    ).visit_with_conv(tp, conversion)
    json_schema = full_schema(json_schema, schema)
    if add_defs and version.defs:
        defs = _refs_schema(
            builder, default_conversion, refs, ref_factory, additional_properties
        )
        if defs:
            json_schema["$defs"] = defs
    result = serialize(
        JsonSchema,
        json_schema,
        aliaser=aliaser,
        check_type=True,
        conversion=version.conversion,
        default_conversion=converters.default_serialization,
        fall_back_on_any=True,
    )
    if with_schema and version.schema is not None:
        result["$schema"] = version.schema
    return result


@deprecate_kwargs({"conversions": "conversion"})
def deserialization_schema(
    tp: AnyType,
    *,
    additional_properties: bool = None,
    aliaser: Aliaser = None,
    all_refs: bool = None,
    conversion: AnyConversion = None,
    default_conversion: DefaultConversion = None,
    ref_factory: RefFactory = None,
    schema: Schema = None,
    version: JsonSchemaVersion = None,
    with_schema: bool = True,
) -> Mapping[str, Any]:
    from apischema import settings

    return _schema(
        DeserializationSchemaBuilder,
        tp,
        schema,
        conversion,
        default_conversion or settings.deserialization.default_conversion,
        version,
        aliaser,
        ref_factory,
        all_refs,
        with_schema,
        additional_properties,
    )


@deprecate_kwargs({"conversions": "conversion"})
def serialization_schema(
    tp: AnyType,
    *,
    additional_properties: bool = None,
    all_refs: bool = None,
    aliaser: Aliaser = None,
    conversion: AnyConversion = None,
    default_conversion: DefaultConversion = None,
    ref_factory: RefFactory = None,
    schema: Schema = None,
    version: JsonSchemaVersion = None,
    with_schema: bool = True,
) -> Mapping[str, Any]:
    from apischema import settings

    return _schema(
        SerializationSchemaBuilder,
        tp,
        schema,
        conversion,
        default_conversion or settings.serialization.default_conversion,
        version,
        aliaser,
        ref_factory,
        all_refs,
        with_schema,
        additional_properties,
    )


def _defs_schema(
    types: TypesWithConversion,
    default_conversion: DefaultConversion,
    builder: Type[SchemaBuilder],
    ref_factory: RefFactory,
    all_refs: bool,
    additional_properties: bool,
) -> Mapping[str, JsonSchema]:
    return _refs_schema(
        builder,
        default_conversion,
        _extract_refs(types, default_conversion, builder, all_refs),
        ref_factory,
        additional_properties,
    )


def _set_missing_properties(
    schema: JsonSchema, properties: Optional[Mapping[str, JsonSchema]], key: str
) -> JsonSchema:
    if properties is None:
        return schema
    missing = {name: prop for name, prop in properties.items() if prop.get(key, False)}
    schema.setdefault("properties", {}).update(missing)
    return schema


def compare_schemas(write: Any, read: Any) -> Any:
    if isinstance(write, Mapping):
        if not isinstance(read, Mapping):
            raise ValueError
        merged: Dict[str, Any] = {}
        for key in write.keys() | read.keys():
            if key in write and key in read:
                if key == "properties":
                    merged[key] = {}
                    for prop in write[key].keys() | read[key].keys():
                        if prop in write[key] and prop in read[key]:
                            merged[key][prop] = compare_schemas(
                                write[key][prop], read[key][prop]
                            )
                        elif prop in write[key]:
                            merged[key][prop] = {**write[key][prop], "writeOnly": True}
                        else:
                            merged[key][prop] = {**read[key][prop], "readOnly": True}
                elif key in {
                    "required",
                    "dependentRequired",
                    "additionalProperties",
                    "patternProperties",
                }:
                    merged[key] = write[key]
                else:
                    merged[key] = compare_schemas(write[key], read[key])
            else:
                merged[key] = write.get(key, read.get(key))
        return merged
    elif isinstance(read, Sequence) and not isinstance(read, str):
        if not isinstance(read, Sequence) or len(write) != len(read):
            raise ValueError
        return [compare_schemas(write[i], read[i]) for i in range(len(write))]
    else:
        if not write == read:
            raise ValueError
        return write


def definitions_schema(
    *,
    deserialization: TypesWithConversion = (),
    serialization: TypesWithConversion = (),
    default_deserialization: DefaultConversion = None,
    default_serialization: DefaultConversion = None,
    aliaser: Aliaser = None,
    version: JsonSchemaVersion = None,
    ref_factory: Optional[RefFactory] = None,
    all_refs: bool = None,
    additional_properties: bool = None,
) -> Mapping[str, Mapping[str, Any]]:
    from apischema import settings

    if additional_properties is None:
        additional_properties = settings.additional_properties
    if aliaser is None:
        aliaser = settings.aliaser
    if default_deserialization is None:
        default_deserialization = settings.deserialization.default_conversion
    if default_serialization is None:
        default_serialization = settings.serialization.default_conversion
    version, ref_factory, all_refs = _default_version(version, ref_factory, all_refs)
    deserialization_schemas = _defs_schema(
        deserialization,
        default_deserialization,
        DeserializationSchemaBuilder,
        ref_factory,
        all_refs,
        additional_properties,
    )
    serialization_schemas = _defs_schema(
        serialization,
        default_serialization,
        SerializationSchemaBuilder,
        ref_factory,
        all_refs,
        additional_properties,
    )
    schemas = {}
    for ref in deserialization_schemas.keys() | serialization_schemas.keys():
        if ref in deserialization_schemas and ref in serialization_schemas:
            try:
                schemas[ref] = compare_schemas(
                    deserialization_schemas[ref], serialization_schemas[ref]
                )
            except ValueError:
                raise TypeError(
                    f"Reference {ref} has different schemas"
                    f" for deserialization and serialization"
                )
        else:
            schemas[ref] = deserialization_schemas.get(
                ref, serialization_schemas.get(ref)
            )
    return {
        ref: serialize(
            JsonSchema,
            schema,
            aliaser=aliaser,
            fall_back_on_any=True,
            check_type=True,
            conversion=version.conversion,
            default_conversion=converters.default_serialization,
        )
        for ref, schema in schemas.items()
    }



from base64 import b64decode, b64encode


def decode_base_64(s: str) -> str:
    return b64decode(s).decode()


def encode_base64(s: str) -> str:
    return b64encode(s.encode()).decode()


base64_encoding = (decode_base_64, encode_base64)



__all__ = ["fields_set", "is_set", "set_fields", "unset_fields", "with_fields_set"]
from dataclasses import (  # type: ignore
    Field,
    _FIELD,
    _FIELDS,
    _FIELD_INITVAR,
    is_dataclass,
)
from functools import wraps
from inspect import signature
from typing import AbstractSet, Any, Collection, Set, Type, TypeVar, cast

from apischema.objects.fields import get_field_name
from apischema.utils import PREFIX

FIELDS_SET_ATTR = f"{PREFIX}fields_set"
_ALREADY_SET = f"{PREFIX}already_set"

Cls = TypeVar("Cls", bound=Type)

_fields_set_classes: Set[type] = set()


def support_fields_set(cls: type) -> bool:
    return any(base in _fields_set_classes for base in cls.__mro__)


def with_fields_set(cls: Cls) -> Cls:
    from apischema.metadata.keys import DEFAULT_AS_SET_METADATA

    init_fields = set()
    post_init_fields = set()
    if is_dataclass(cls):
        for field in getattr(cls, _FIELDS).values():
            assert isinstance(field, Field)
            if field._field_type == _FIELD_INITVAR:  # type: ignore
                init_fields.add(field.name)
            if field._field_type == _FIELD and not field.init:  # type: ignore
                post_init_fields.add(field.name)
            if field.metadata.get(DEFAULT_AS_SET_METADATA):
                post_init_fields.add(field.name)
    params = list(signature(cls.__init__).parameters)[1:]
    old_new = cls.__new__
    old_init = cls.__init__
    old_setattr = cls.__setattr__

    def new_new(*args, **kwargs):
        if old_new is object.__new__:
            obj = object.__new__(args[0])
        else:
            obj = old_new(*args, **kwargs)
        # Initialize FIELD_SET_ATTR in order to prevent inherited class which override
        # __init__ to raise in __setattr__
        obj.__dict__[FIELDS_SET_ATTR] = set()
        return obj

    def new_init(self, *args, **kwargs):
        prev_fields_set = self.__dict__.get(FIELDS_SET_ATTR, set()).copy()
        self.__dict__[FIELDS_SET_ATTR] = set()
        try:
            old_init(self, *args, **kwargs)
        except TypeError as err:
            if str(err) == no_dataclass_init_error:
                raise RuntimeError(dataclass_before_error) from None
            else:
                raise
        arg_fields = {*params[: len(args)], *kwargs} - init_fields
        self.__dict__[FIELDS_SET_ATTR] = prev_fields_set | arg_fields | post_init_fields

    def new_setattr(self, attr, value):
        try:
            self.__dict__[FIELDS_SET_ATTR].add(attr)
        except KeyError:
            raise RuntimeError(dataclass_before_error) from None
        old_setattr(self, attr, value)

    for attr, old, new in [
        ("__new__", old_new, new_new),
        ("__init__", old_init, new_init),
        ("__setattr__", old_setattr, new_setattr),
    ]:
        if hasattr(old, _ALREADY_SET):
            continue
        setattr(new, _ALREADY_SET, True)
        setattr(cls, attr, wraps(old)(new))  # type: ignore

    _fields_set_classes.add(cls)
    return cls


no_dataclass_init_error = (
    "object.__init__() takes exactly one argument (the instance to initialize)"
)
dataclass_before_error = (
    f"{with_fields_set.__name__} must be put before dataclass decorator"
)


T = TypeVar("T")


def _field_names(fields: Collection) -> AbstractSet[str]:
    result: Set[str] = set()
    for field in fields:
        result.add(get_field_name(field))
    return result


def _fields_set(obj: Any) -> Set[str]:
    try:
        return getattr(obj, FIELDS_SET_ATTR)
    except AttributeError:
        raise TypeError(
            f"Type {obj.__class__} is not decorated" f" with {with_fields_set.__name__}"
        )


def set_fields(obj: T, *fields: Any, overwrite=False) -> T:
    if overwrite:
        _fields_set(obj).clear()
    _fields_set(obj).update(map(get_field_name, fields))
    return obj


def unset_fields(obj: T, *fields: Any) -> T:
    _fields_set(obj).difference_update(map(get_field_name, fields))
    return obj


# This could just be an alias with a specified type, but it's better handled by IDE
# like this
def fields_set(obj: Any) -> AbstractSet[str]:
    return _fields_set(obj)


class FieldIsSet:
    def __init__(self, obj: Any):
        self.fields_set = fields_set(obj)

    def __getattribute__(self, name: str) -> bool:
        return name in object.__getattribute__(self, "fields_set")


def is_set(obj: T) -> T:
    return cast(T, FieldIsSet(obj))



__all__ = [
    "AliasedStr",
    "ObjectField",
    "get_alias",
    "get_field",
    "object_deserialization",
    "object_fields",
    "object_serialization",
    "set_object_fields",
]
from .conversions import object_deserialization, object_serialization
from .fields import ObjectField, set_object_fields
from .getters import AliasedStr, get_alias, get_field, object_fields



import warnings

from apischema.dependencies import dependent_required


def DependentRequired(*args, **kwargs):
    warnings.warn(
        "apischema.dependent_required.DependentRequired is deprecated,"
        " use apischema.dependent_required instead",
        DeprecationWarning,
    )
    return dependent_required(*args, **kwargs)



import inspect
from dataclasses import Field, replace
from types import new_class
from typing import (
    Any,
    Callable,
    Dict,
    Generic,
    Iterable,
    Mapping,
    Sequence,
    Tuple,
    Type,
    TypeVar,
    Union,
)

from apischema.methods import is_method, method_wrapper
from apischema.objects.fields import MISSING_DEFAULT, ObjectField, set_object_fields
from apischema.objects.getters import object_fields, parameters_as_fields
from apischema.type_names import type_name
from apischema.types import OrderedDict
from apischema.typing import get_type_hints
from apischema.utils import (
    empty_dict,
    substitute_type_vars,
    subtyping_substitution,
    to_pascal_case,
    with_parameters,
)

T = TypeVar("T")


def object_deserialization(
    func: Callable[..., T],
    *input_class_modifiers: Callable[[type], Any],
    parameters_metadata: Mapping[str, Mapping] = None,
) -> Any:
    fields = parameters_as_fields(func, parameters_metadata)
    types = get_type_hints(func, include_extras=True)
    if "return" not in types:
        raise TypeError("Object deserialization must be typed")
    return_type = types["return"]
    bases = ()
    if getattr(return_type, "__parameters__", ()):
        bases = (Generic[return_type.__parameters__],)  # type: ignore
    elif func.__name__ != "<lambda>":
        input_class_modifiers = (
            type_name(to_pascal_case(func.__name__)),
            *input_class_modifiers,
        )

    def __init__(self, **kwargs):
        self.kwargs = kwargs

    input_cls = new_class(
        to_pascal_case(func.__name__),
        bases,
        exec_body=lambda ns: ns.update({"__init__": __init__}),
    )
    for modifier in input_class_modifiers:
        modifier(input_cls)
    set_object_fields(input_cls, fields)
    if any(f.additional_properties for f in fields):
        kwargs_param = next(f.name for f in fields if f.additional_properties)

        def wrapper(input):
            kwargs = input.kwargs.copy()
            kwargs.update(kwargs.pop(kwargs_param))
            return func(**kwargs)

    else:

        def wrapper(input):
            return func(**input.kwargs)

    wrapper.__annotations__["input"] = with_parameters(input_cls)
    wrapper.__annotations__["return"] = return_type
    return wrapper


def _fields_and_init(
    cls: type, fields_and_methods: Union[Iterable[Any], Callable[[], Iterable[Any]]]
) -> Tuple[Sequence[ObjectField], Callable[[Any, Any], None]]:
    fields = object_fields(cls, serialization=True)
    output_fields: Dict[str, ObjectField] = OrderedDict()
    methods = []
    if callable(fields_and_methods):
        fields_and_methods = fields_and_methods()
    for elt in fields_and_methods:
        if elt is ...:
            output_fields.update(fields)
            continue
        if isinstance(elt, tuple):
            elt, metadata = elt
        else:
            metadata = empty_dict
        if not isinstance(metadata, Mapping):
            raise TypeError(f"Invalid metadata {metadata}")
        if isinstance(elt, Field):
            elt = elt.name
        if isinstance(elt, str) and elt in fields:
            elt = fields[elt]
        if is_method(elt):
            elt = method_wrapper(elt)
        if isinstance(elt, ObjectField):
            if metadata:
                output_fields[elt.name] = replace(
                    elt, metadata={**elt.metadata, **metadata}, default=MISSING_DEFAULT
                )
            else:
                output_fields[elt.name] = elt
            continue
        elif callable(elt):
            types = get_type_hints(elt)
            first_param = next(iter(inspect.signature(elt).parameters))
            substitution, _ = subtyping_substitution(types.get(first_param, cls), cls)
            ret = substitute_type_vars(types.get("return", Any), substitution)
            output_fields[elt.__name__] = ObjectField(
                elt.__name__, ret, metadata=metadata
            )
            methods.append((elt, output_fields[elt.__name__]))
        else:
            raise TypeError(f"Invalid serialization member {elt} for class {cls}")

    serialized_methods = [m for m, f in methods if output_fields[f.name] is f]
    serialized_fields = list(
        output_fields.keys() - {m.__name__ for m in serialized_methods}
    )

    def __init__(self, obj):
        for field in serialized_fields:
            setattr(self, field, getattr(obj, field))
        for method in serialized_methods:
            setattr(self, method.__name__, method(obj))

    return tuple(output_fields.values()), __init__


def object_serialization(
    cls: Type[T],
    fields_and_methods: Union[Iterable[Any], Callable[[], Iterable[Any]]],
    *output_class_modifiers: Callable[[type], Any],
) -> Callable[[T], Any]:

    generic, bases = cls, ()
    if getattr(cls, "__parameters__", ()):
        generic = cls[cls.__parameters__]  # type: ignore
        bases = Generic[cls.__parameters__]  # type: ignore
    elif (
        callable(fields_and_methods)
        and fields_and_methods.__name__ != "<lambda>"
        and not getattr(cls, "__parameters__", ())
    ):
        output_class_modifiers = (
            type_name(to_pascal_case(fields_and_methods.__name__)),
            *output_class_modifiers,
        )

    def __init__(self, obj):
        _, new_init = _fields_and_init(cls, fields_and_methods)
        new_init.__annotations__ = {"obj": generic}
        output_cls.__init__ = new_init
        new_init(self, obj)

    __init__.__annotations__ = {"obj": generic}
    output_cls = new_class(
        f"{cls.__name__}Serialization",
        bases,
        exec_body=lambda ns: ns.update({"__init__": __init__}),
    )
    for modifier in output_class_modifiers:
        modifier(output_cls)
    set_object_fields(output_cls, lambda: _fields_and_init(cls, fields_and_methods)[0])

    return output_cls



from dataclasses import dataclass
from typing import Any, Callable, ClassVar, Dict, Optional

from apischema.conversions import Conversion, LazyConversion
from apischema.json_schema.types import JsonSchema, JsonType

RefFactory = Callable[[str], str]


def ref_prefix(prefix: str) -> RefFactory:
    if not prefix.endswith("/"):
        prefix += "/"
    return lambda ref: prefix + ref


def isolate_ref(schema: Dict[str, Any]):
    if "$ref" in schema and len(schema) > 1:
        schema.setdefault("allOf", []).append({"$ref": schema.pop("$ref")})


def to_json_schema_2019_09(schema: JsonSchema) -> Dict[str, Any]:
    result = schema.copy()
    if "prefixItems" in result:
        if "items" in result:
            result["additionalItems"] = result.pop("items")
        result["items"] = result["prefixItems"]
    return result


def to_json_schema_7(schema: JsonSchema) -> Dict[str, Any]:
    result = to_json_schema_2019_09(schema)
    isolate_ref(result)
    if "$defs" in result:
        result["definitions"] = {**result.pop("$defs"), **result.get("definitions", {})}
    if "dependentRequired" in result:
        result["dependencies"] = {
            **result.pop("dependentRequired"),
            **result.get("dependencies", {}),
        }
    return result


OPEN_API_3_0_UNSUPPORTED = [
    "dependentRequired",
    "unevaluatedProperties",
    "additionalItems",
]


def to_open_api_3_0(schema: JsonSchema) -> Dict[str, Any]:
    result = to_json_schema_2019_09(schema)
    for key in OPEN_API_3_0_UNSUPPORTED:
        result.pop(key, ...)
    isolate_ref(result)
    if {"type": "null"} in result.get("anyOf", ()):
        result.setdefault("nullable", True)
        result["anyOf"] = [a for a in result["anyOf"] if a != {"type": "null"}]
    if "type" in result and not isinstance(result["type"], (str, JsonType)):
        if "null" in result["type"]:
            result.setdefault("nullable", True)
        result["type"] = [t for t in result["type"] if t != "null"]
        if len(result["type"]) > 1:
            result.setdefault("anyOf", []).extend(
                {"type": t} for t in result.pop("type")
            )
        else:
            result["type"] = result["type"][0]
    if "examples" in result:
        result.setdefault("example", result.pop("examples")[0])
    if "const" in result:
        result.setdefault("enum", [result.pop("const")])
    return result


@dataclass
class JsonSchemaVersion:
    schema: Optional[str] = None
    ref_prefix: str = ""
    serialization: Optional[Callable] = None
    all_refs: bool = True
    defs: bool = True

    @property
    def conversion(self) -> Optional[Conversion]:
        if self.serialization:
            # Recursive conversion pattern
            tmp = None
            conversion = Conversion(
                self.serialization, sub_conversion=LazyConversion(lambda: tmp)
            )
            tmp = conversion
            return conversion
        else:
            return None

    @property
    def ref_factory(self) -> RefFactory:
        return ref_prefix(self.ref_prefix)

    DRAFT_2020_12: ClassVar["JsonSchemaVersion"]
    DRAFT_2019_09: ClassVar["JsonSchemaVersion"]
    DRAFT_7: ClassVar["JsonSchemaVersion"]
    OPEN_API_3_0: ClassVar["JsonSchemaVersion"]
    OPEN_API_3_1: ClassVar["JsonSchemaVersion"]


JsonSchemaVersion.DRAFT_2020_12 = JsonSchemaVersion(
    "http://json-schema.org/draft/2020-12/schema#", "#/$defs/", None, False, True
)
JsonSchemaVersion.DRAFT_2019_09 = JsonSchemaVersion(
    "http://json-schema.org/draft/2020-12/schema#",
    "#/$defs/",
    to_json_schema_2019_09,
    False,
    True,
)
JsonSchemaVersion.DRAFT_7 = JsonSchemaVersion(
    "http://json-schema.org/draft-07/schema#",
    "#/definitions/",
    to_json_schema_7,
    False,
    True,
)
JsonSchemaVersion.OPEN_API_3_0 = JsonSchemaVersion(
    None, "#/components/schemas/", to_open_api_3_0, True, False
)
JsonSchemaVersion.OPEN_API_3_1 = JsonSchemaVersion(
    None, "#/components/schemas/", None, True, False
)



from dataclasses import Field, InitVar, MISSING, dataclass, field
from enum import Enum, auto
from types import FunctionType
from typing import (
    Any,
    Callable,
    Iterable,
    Mapping,
    MutableMapping,
    NoReturn,
    Optional,
    Pattern,
    Sequence,
    TYPE_CHECKING,
    Union,
    cast,
)

from apischema.cache import CacheAwareDict
from apischema.conversions.conversions import AnyConversion
from apischema.metadata.implem import (
    ConversionMetadata,
    SkipMetadata,
    ValidatorsMetadata,
)
from apischema.metadata.keys import (
    ALIAS_METADATA,
    ALIAS_NO_OVERRIDE_METADATA,
    CONVERSION_METADATA,
    DEFAULT_AS_SET_METADATA,
    FALL_BACK_ON_DEFAULT_METADATA,
    FLATTEN_METADATA,
    NONE_AS_UNDEFINED_METADATA,
    ORDERING_METADATA,
    POST_INIT_METADATA,
    PROPERTIES_METADATA,
    REQUIRED_METADATA,
    SCHEMA_METADATA,
    SKIP_METADATA,
    VALIDATORS_METADATA,
)
from apischema.types import AnyType, ChainMap, NoneType, UndefinedType
from apischema.typing import get_args, is_annotated
from apischema.utils import (
    LazyValue,
    empty_dict,
    get_args2,
    is_union_of,
    keep_annotations,
)

if TYPE_CHECKING:
    from apischema.ordering import Ordering
    from apischema.schemas import Schema
    from apischema.validation.validators import Validator


class FieldKind(Enum):
    NORMAL = auto()
    READ_ONLY = auto()
    WRITE_ONLY = auto()


# Cannot reuse MISSING for dataclass field because it would be interpreted as no default
MISSING_DEFAULT = object()


@dataclass(frozen=True)
class ObjectField:
    name: str
    type: AnyType
    required: bool = True
    metadata: Mapping[str, Any] = field(default_factory=lambda: empty_dict)
    default: InitVar[Any] = MISSING_DEFAULT
    default_factory: Optional[Callable[[], Any]] = None
    kind: FieldKind = FieldKind.NORMAL

    def __post_init__(self, default: Any):
        if REQUIRED_METADATA in self.full_metadata:
            object.__setattr__(self, "required", True)
        if self.default_factory is MISSING:
            object.__setattr__(self, "default_factory", None)
        if not self.required and self.default_factory is None:
            if default is MISSING_DEFAULT:
                raise ValueError("Missing default for non-required ObjectField")
            object.__setattr__(self, "default_factory", LazyValue(default))
        if self.none_as_undefined and is_union_of(self.type, NoneType):
            new_type = Union[tuple(a for a in get_args2(self.type) if a != NoneType)]  # type: ignore
            object.__setattr__(self, "type", keep_annotations(new_type, self.type))

    @property
    def full_metadata(self) -> Mapping[str, Any]:
        if not is_annotated(self.type):
            return self.metadata
        return ChainMap(
            self.metadata,
            *(
                arg
                for arg in reversed(get_args(self.type)[1:])
                if isinstance(arg, Mapping)
            ),
        )

    @property
    def additional_properties(self) -> bool:
        return self.full_metadata.get(PROPERTIES_METADATA, ...) is None

    @property
    def alias(self) -> str:
        return self.full_metadata.get(ALIAS_METADATA, self.name)

    @property
    def override_alias(self) -> bool:
        return ALIAS_NO_OVERRIDE_METADATA not in self.full_metadata

    @property
    def _conversion(self) -> Optional[ConversionMetadata]:
        return self.metadata.get(CONVERSION_METADATA)

    @property
    def default_as_set(self) -> bool:
        return DEFAULT_AS_SET_METADATA in self.full_metadata

    @property
    def deserialization(self) -> Optional[AnyConversion]:
        conversion = self._conversion
        return conversion.deserialization if conversion is not None else None

    @property
    def fall_back_on_default(self) -> bool:
        return (
            FALL_BACK_ON_DEFAULT_METADATA in self.full_metadata
            and self.default_factory is not None
        )

    @property
    def flattened(self) -> bool:
        return FLATTEN_METADATA in self.full_metadata

    def get_default(self) -> Any:
        if self.required:
            raise RuntimeError("Field is required")
        assert self.default_factory is not None
        return self.default_factory()  # type: ignore

    @property
    def is_aggregate(self) -> bool:
        return (
            self.flattened
            or self.additional_properties
            or self.pattern_properties is not None
        )

    @property
    def none_as_undefined(self):
        return NONE_AS_UNDEFINED_METADATA in self.full_metadata

    @property
    def ordering(self) -> Optional["Ordering"]:
        return self.full_metadata.get(ORDERING_METADATA)

    @property
    def post_init(self) -> bool:
        return POST_INIT_METADATA in self.full_metadata

    @property
    def pattern_properties(self) -> Union[Pattern, "ellipsis", None]:  # noqa: F821
        return self.full_metadata.get(PROPERTIES_METADATA)

    @property
    def schema(self) -> Optional["Schema"]:
        return self.metadata.get(SCHEMA_METADATA)

    @property
    def serialization(self) -> Optional[AnyConversion]:
        conversion = self._conversion
        return conversion.serialization if conversion is not None else None

    @property
    def skip(self) -> SkipMetadata:
        return self.metadata.get(SKIP_METADATA, SkipMetadata())

    def skippable(self, default: bool, none: bool) -> bool:
        return bool(
            self.skip.serialization_if
            or is_union_of(self.type, UndefinedType)
            or (
                self.default_factory is not None
                and (self.skip.serialization_default or default)
            )
            or self.none_as_undefined
            or (none and is_union_of(self.type, NoneType))
        )

    @property
    def undefined(self) -> bool:
        return is_union_of(self.type, UndefinedType)

    @property
    def validators(self) -> Sequence["Validator"]:
        if VALIDATORS_METADATA in self.metadata:
            return cast(
                ValidatorsMetadata, self.metadata[VALIDATORS_METADATA]
            ).validators
        else:
            return ()


FieldOrName = Union[str, ObjectField, Field]


def _bad_field(obj: Any, methods: bool) -> NoReturn:
    method_types = "property/types.FunctionType" if methods else ""
    raise TypeError(
        f"Expected dataclasses.Field/apischema.ObjectField/str{method_types}, found {obj}"
    )


def check_field_or_name(field_or_name: Any, *, methods: bool = False):
    method_types = (property, FunctionType) if methods else ()
    if not isinstance(field_or_name, (str, ObjectField, Field, *method_types)):
        _bad_field(field_or_name, methods)


def get_field_name(field_or_name: Any, *, methods: bool = False) -> str:
    if isinstance(field_or_name, (Field, ObjectField)):
        return field_or_name.name
    elif isinstance(field_or_name, str):
        return field_or_name
    elif (
        methods
        and isinstance(field_or_name, property)
        and field_or_name.fget is not None
    ):
        return field_or_name.fget.__name__
    elif methods and isinstance(field_or_name, FunctionType):
        return field_or_name.__name__
    else:
        _bad_field(field_or_name, methods)


_class_fields: MutableMapping[
    type, Callable[[], Sequence[ObjectField]]
] = CacheAwareDict({})


def set_object_fields(
    cls: type,
    fields: Union[Iterable[ObjectField], Callable[[], Sequence[ObjectField]], None],
):
    if fields is None:
        _class_fields.pop(cls, ...)
    elif callable(fields):
        _class_fields[cls] = fields
    else:
        _class_fields[cls] = lambda fields=tuple(fields): fields  # type: ignore


def default_object_fields(cls: type) -> Optional[Sequence[ObjectField]]:
    return _class_fields[cls]() if cls in _class_fields else None



__all__ = [
    "AnyConversion",
    "Conversion",
    "LazyConversion",
    "as_names",
    "as_str",
    "dataclass_input_wrapper",
    "deserializer",
    "reset_deserializers",
    "reset_serializer",
    "serializer",
]

import sys
import warnings

from .conversions import AnyConversion, Conversion, LazyConversion
from .converters import (
    as_names,
    as_str,
    deserializer,
    reset_deserializers,
    reset_serializer,
    serializer,
)
from .wrappers import dataclass_input_wrapper

if sys.version_info >= (3, 7):

    def __getattr__(name):
        if name == "identity":
            from apischema.utils import identity  # noqa: F811

            warnings.warn(
                "apischema.conversions.identity is deprecated, "
                "use apischema.identity instead",
                DeprecationWarning,
            )
            return identity
        raise AttributeError(f"module {__name__} has no attribute {name}")


else:
    from apischema.utils import identity  # noqa: F401



from collections import defaultdict
from functools import wraps
from inspect import Parameter, isgeneratorfunction, signature
from itertools import chain
from types import MethodType
from typing import (
    AbstractSet,
    Any,
    Callable,
    Collection,
    Iterable,
    List,
    Mapping,
    MutableMapping,
    Optional,
    Sequence,
    Type,
    TypeVar,
    overload,
)

from apischema.aliases import Aliaser
from apischema.cache import CacheAwareDict
from apischema.methods import is_method, method_class
from apischema.objects import get_alias
from apischema.objects.fields import FieldOrName, check_field_or_name, get_field_name
from apischema.types import AnyType
from apischema.typing import get_type_hints
from apischema.utils import get_origin_or_type2
from apischema.validation.dependencies import find_all_dependencies
from apischema.validation.errors import (
    ValidationError,
    apply_aliaser,
    build_validation_error,
    merge_errors,
)
from apischema.validation.mock import NonTrivialDependency

_validators: MutableMapping[Type, List["Validator"]] = CacheAwareDict(defaultdict(list))


def get_validators(tp: AnyType) -> Sequence["Validator"]:
    return list(
        chain.from_iterable(_validators[cls] for cls in getattr(tp, "__mro__", [tp]))
    )


class Discard(Exception):
    def __init__(self, fields: Optional[AbstractSet[str]], error: ValidationError):
        self.fields = fields
        self.error = error


class Validator:
    def __init__(
        self,
        func: Callable,
        field: FieldOrName = None,
        discard: Collection[FieldOrName] = None,
    ):
        wraps(func)(self)
        self.func = func
        self.field = field
        # Cannot use field.name because fields are not yet initialized with __set_name__
        if field is not None and discard is None:
            self.discard: Optional[Collection[FieldOrName]] = (field,)
        else:
            self.discard = discard
        self.dependencies: AbstractSet[str] = set()
        try:
            parameters = signature(func).parameters
        except ValueError:
            self.params: AbstractSet[str] = set()
        else:
            if not parameters:
                raise TypeError("Validator must have at least one parameter")
            if any(p.kind == Parameter.VAR_KEYWORD for p in parameters.values()):
                raise TypeError("Validator cannot have variadic keyword parameter")
            if any(p.kind == Parameter.VAR_POSITIONAL for p in parameters.values()):
                raise TypeError("Validator cannot have variadic positional parameter")
            self.params = set(list(parameters)[1:])
        if isgeneratorfunction(func):

            def validate(*args, **kwargs):
                errors = list(func(*args, **kwargs))
                if errors:
                    raise build_validation_error(errors)

            self.validate = validate

        else:
            self.validate = func

    def __get__(self, instance, owner):
        return self if instance is None else MethodType(self.func, instance)

    def __call__(self, *args, **kwargs):
        raise RuntimeError("Method __set_name__ has not been called")

    def _register(self, owner: Type):
        self.owner = owner
        self.dependencies = find_all_dependencies(owner, self.func) | self.params
        _validators[owner].append(self)

    def __set_name__(self, owner, name):
        self._register(owner)
        setattr(owner, name, self.func)


T = TypeVar("T")


def validate(
    obj: T,
    validators: Iterable[Validator] = None,
    kwargs: Optional[Mapping[str, Any]] = None,
    *,
    aliaser: Aliaser = lambda s: s,
) -> T:
    if validators is None:
        validators = get_validators(obj.__class__)
    else:
        validators = list(validators)
    error: Optional[ValidationError] = None
    for i, validator in enumerate(validators):
        try:
            if not kwargs:
                validator.validate(obj)
            elif validator.params == kwargs.keys():
                validator.validate(obj, **kwargs)
            else:
                validator.validate(obj, **{k: kwargs[k] for k in validator.params})
        except ValidationError as e:
            err = apply_aliaser(e, aliaser)
        except NonTrivialDependency as exc:
            exc.validator = validator
            raise
        except AssertionError:
            raise
        except Exception as e:
            err = ValidationError([str(e)])
        else:
            continue
        if validator.field is not None:
            alias = getattr(get_alias(validator.owner), get_field_name(validator.field))
            err = ValidationError(children={aliaser(alias): err})
        error = merge_errors(error, err)
        if validator.discard:
            try:
                discarded = set(map(get_field_name, validator.discard))
                next_validators = (
                    v for v in validators[i:] if v.dependencies.isdisjoint(discarded)
                )
                validate(obj, next_validators, kwargs, aliaser=aliaser)
            except ValidationError as err:
                raise merge_errors(error, err)
            else:
                raise error
    if error is not None:
        raise error
    return obj


V = TypeVar("V", bound=Callable)


@overload
def validator(func: V) -> V:
    ...


@overload
def validator(
    field: Any = None, *, discard: Any = None, owner: Type = None
) -> Callable[[V], V]:
    ...


def validator(arg=None, *, field=None, discard=None, owner=None):
    if callable(arg):
        validator_ = Validator(arg, field, discard)
        if is_method(arg):
            cls = method_class(arg)
            if cls is None:
                if owner is not None:
                    raise TypeError("Validator owner cannot be set for class validator")
                return validator_
            elif owner is None:
                owner = cls
        if owner is None:
            try:
                first_param = next(iter(signature(arg).parameters))
                owner = get_origin_or_type2(get_type_hints(arg)[first_param])
            except Exception:
                raise ValueError("Validator first parameter must be typed")
        validator_._register(owner)
        return arg
    else:
        field = field or arg
        if field is not None:
            check_field_or_name(field)
        if discard is not None:
            if not isinstance(discard, Collection) or isinstance(discard, str):
                discard = [discard]
            for discarded in discard:
                check_field_or_name(discarded)
        return lambda func: validator(func, field=field, discard=discard, owner=owner)



__all__ = ["cache", "reset", "set_size"]
import sys
from functools import lru_cache
from typing import Callable, Iterator, MutableMapping, TypeVar, cast

from apischema.utils import type_dict_wrapper

_cached: list = []

Func = TypeVar("Func", bound=Callable)


def cache(func: Func) -> Func:
    cached = cast(Func, lru_cache()(func))
    _cached.append(cached)
    return cached


def reset():
    for cached in _cached:
        cached.cache_clear()


def set_size(size: int):
    for cached in _cached:
        wrapped = cached.__wrapped__
        setattr(
            sys.modules[wrapped.__module__], wrapped.__name__, lru_cache(size)(wrapped)
        )


K = TypeVar("K")
V = TypeVar("V")


class CacheAwareDict(MutableMapping[K, V]):
    def __init__(self, wrapped: MutableMapping[K, V]):
        self.wrapped = type_dict_wrapper(wrapped)

    def __getitem__(self, key: K) -> V:
        return self.wrapped[key]

    def __setitem__(self, key: K, value: V):
        self.wrapped[key] = value
        reset()

    def __delitem__(self, key: K):
        del self.wrapped[key]

    def __len__(self) -> int:
        return len(self.wrapped)

    def __iter__(self) -> Iterator[K]:
        return iter(self.wrapped)



from collections import defaultdict
from dataclasses import dataclass
from typing import (
    Any,
    Callable,
    Collection,
    Dict,
    List,
    Mapping,
    MutableMapping,
    Optional,
    Sequence,
    TypeVar,
    overload,
)

from apischema.cache import CacheAwareDict
from apischema.metadata.keys import ORDERING_METADATA
from apischema.types import MetadataMixin
from apischema.utils import stop_signature_abuse

Cls = TypeVar("Cls", bound=type)


@dataclass(frozen=True)
class Ordering(MetadataMixin):
    key = ORDERING_METADATA
    order: Optional[int] = None
    after: Optional[Any] = None
    before: Optional[Any] = None

    def __post_init__(self):
        from apischema.objects.fields import check_field_or_name

        if self.after is not None:
            check_field_or_name(self.after, methods=True)
        if self.before is not None:
            check_field_or_name(self.before, methods=True)


_order_overriding: MutableMapping[type, Mapping[Any, Ordering]] = CacheAwareDict({})


@overload
def order(__value: int) -> Ordering:
    ...


@overload
def order(*, after: Any) -> Ordering:
    ...


@overload
def order(*, before: Any) -> Ordering:
    ...


@overload
def order(__fields: Sequence[Any]) -> Callable[[Cls], Cls]:
    ...


@overload
def order(__override: Mapping[Any, Ordering]) -> Callable[[Cls], Cls]:
    ...


def order(__arg=None, *, before=None, after=None):
    if len([arg for arg in (__arg, before, after) if arg is not None]) != 1:
        stop_signature_abuse()
    if isinstance(__arg, Sequence):
        __arg = {field: order(after=prev) for field, prev in zip(__arg[1:], __arg)}
    if isinstance(__arg, Mapping):
        if not all(isinstance(val, Ordering) for val in __arg.values()):
            stop_signature_abuse()

        def decorator(cls: Cls) -> Cls:
            _order_overriding[cls] = __arg
            return cls

        return decorator
    elif __arg is not None and not isinstance(__arg, int):
        stop_signature_abuse()
    else:
        return Ordering(__arg, after, before)


def get_order_overriding(cls: type) -> Mapping[str, Ordering]:
    from apischema.objects.fields import get_field_name

    return {
        get_field_name(field, methods=True): ordering
        for sub_cls in reversed(cls.__mro__)
        if sub_cls in _order_overriding
        for field, ordering in _order_overriding[sub_cls].items()
    }


T = TypeVar("T")


def sort_by_order(
    cls: type,
    elts: Collection[T],
    name: Callable[[T], str],
    order: Callable[[T], Optional[Ordering]],
) -> Sequence[T]:
    from apischema.objects.fields import get_field_name

    order_overriding = get_order_overriding(cls)
    groups: Dict[int, List[T]] = defaultdict(list)
    after: Dict[str, List[T]] = defaultdict(list)
    before: Dict[str, List[T]] = defaultdict(list)
    for elt in elts:
        ordering = order_overriding.get(name(elt), order(elt))
        if ordering is None:
            groups[0].append(elt)
        elif ordering.order is not None:
            groups[ordering.order].append(elt)
        elif ordering.after is not None:
            after[get_field_name(ordering.after, methods=True)].append(elt)
        elif ordering.before is not None:
            before[get_field_name(ordering.before, methods=True)].append(elt)
        else:
            raise NotImplementedError
    if not after and not before and len(groups) == 1:
        return next(iter(groups.values()))
    result = []

    def add_to_result(elt: T):
        elt_name = name(elt)
        for before_elt in before[elt_name]:
            add_to_result(before_elt)
        result.append(elt)
        for after_elt in after[elt_name]:
            add_to_result(after_elt)

    for value in sorted(groups):
        for elt in groups[value]:
            add_to_result(elt)
    return result



from dataclasses import dataclass, field, replace
from functools import reduce
from typing import (
    Any,
    Collection,
    Dict,
    Generator,
    Iterable,
    Iterator,
    List,
    Mapping,
    Optional,
    Sequence,
    Tuple,
    TypeVar,
    Union,
    overload,
)

from apischema.aliases import Aliaser
from apischema.objects import AliasedStr
from apischema.utils import merge_opts

try:
    from apischema.typing import Annotated
except ImportError:
    Annotated = ...  # type: ignore

ErrorMsg = str
Error = Union[ErrorMsg, Tuple[Any, ErrorMsg]]
# where  Any = Union[Field, int, str, Iterable[Union[Field, int, str,]]]
# but Field being kind of magic not understood by type checkers, it's hidden behind Any
ErrorKey = Union[str, int]
T = TypeVar("T")
ValidatorResult = Generator[Error, None, T]

try:
    from apischema.typing import TypedDict

    class LocalizedError(TypedDict):
        loc: Sequence[ErrorKey]
        msg: ErrorMsg


except ImportError:
    LocalizedError = Mapping[str, Any]  # type: ignore


@dataclass
class ValidationError(Exception):
    messages: Sequence[ErrorMsg] = field(default_factory=list)
    children: Mapping[ErrorKey, "ValidationError"] = field(default_factory=dict)

    def __str__(self):
        return repr(self)

    def _errors(self) -> Iterator[Tuple[List[ErrorKey], ErrorMsg]]:
        for msg in self.messages:
            yield [], msg
        for child_key in sorted(self.children):
            for path, error in self.children[child_key]._errors():
                yield [child_key, *path], error

    @property
    def errors(self) -> List[LocalizedError]:
        return [{"loc": path, "msg": error} for path, error in self._errors()]

    @staticmethod
    def from_errors(errors: Sequence[LocalizedError]) -> "ValidationError":
        return reduce(
            merge_errors,
            [_rec_build_error(err["loc"], err["msg"]) for err in errors],
            ValidationError(),
        )


@overload
def merge_errors(
    err1: Optional[ValidationError], err2: ValidationError
) -> ValidationError:
    ...


@overload
def merge_errors(
    err1: ValidationError, err2: Optional[ValidationError]
) -> ValidationError:
    ...


@overload
def merge_errors(
    err1: Optional[ValidationError], err2: Optional[ValidationError]
) -> Optional[ValidationError]:
    ...


@merge_opts  # type: ignore
def merge_errors(err1: ValidationError, err2: ValidationError) -> ValidationError:
    if err1 is None:
        return err2
    if err2 is None:
        return err1
    return ValidationError(
        [*err1.messages, *err2.messages],
        {
            key: merge_errors(  # type: ignore
                err1.children.get(key), err2.children.get(key)
            )
            for key in err1.children.keys() | err2.children.keys()
        },
    )


def apply_aliaser(error: ValidationError, aliaser: Aliaser) -> ValidationError:
    aliased, aliased_children = False, {}
    for key, child in error.children.items():
        if isinstance(key, AliasedStr):
            key = str(aliaser(key))  # str because it could be a str subclass
            aliased = True
        child2 = apply_aliaser(child, aliaser)
        aliased |= child2 is not child
        aliased_children[key] = child2
    return replace(error, children=aliased_children) if aliased else error


def _rec_build_error(path: Sequence[ErrorKey], msg: ErrorMsg) -> ValidationError:
    if not path:
        return ValidationError([msg])
    else:
        return ValidationError(children={path[0]: _rec_build_error(path[1:], msg)})


def build_validation_error(errors: Iterable[Error]) -> ValidationError:
    messages: List[ErrorMsg] = []
    children: Dict[ErrorKey, ValidationError] = {}
    for error in errors:
        if isinstance(error, ErrorMsg):
            messages.append(error)
            continue
        path, msg = error
        if not path:
            messages.append(msg)
        else:
            if isinstance(path, str) or not isinstance(path, Collection):
                path = (path,)
            key, *remain = path
            children[key] = merge_errors(
                children.get(key), _rec_build_error(remain, msg)
            )
    return ValidationError(messages, children)



__all__ = ["Tagged", "TaggedUnion", "get_tagged"]

import warnings
from dataclasses import InitVar, dataclass, field
from typing import (
    Any,
    ClassVar,
    Generic,
    Optional,
    Tuple,
    Type,
    TypeVar,
    Union,
    overload,
)

from apischema.aliases import alias as alias_metadata
from apischema.conversions.conversions import ConvOrFunc
from apischema.metadata import conversion
from apischema.metadata.keys import (
    DEFAULT_AS_SET_METADATA,
    FALL_BACK_ON_DEFAULT_METADATA,
    FLATTEN_METADATA,
    INIT_VAR_METADATA,
    POST_INIT_METADATA,
    PROPERTIES_METADATA,
    REQUIRED_METADATA,
    SKIP_METADATA,
)
from apischema.schemas import Schema, schema
from apischema.types import Metadata, MetadataImplem, Undefined, UndefinedType
from apischema.typing import get_type_hints
from apischema.utils import PREFIX, get_args2, get_origin2, wrap_generic_init_subclass

TAGS_ATTR = f"{PREFIX}tags"

T = TypeVar("T", bound="TaggedUnion")
V = TypeVar("V")


class Tag(str, Generic[T, V]):
    def __new__(cls, tag: str, type: Type[T]):
        return super().__new__(cls, tag)

    def __init__(self, tag: str, type: Type[T]):
        super().__init__()
        self.type = type

    def __call__(self, value: V) -> T:
        return self.type(**{self: value})  # type: ignore


INVALID_METADATA = {
    DEFAULT_AS_SET_METADATA,
    FALL_BACK_ON_DEFAULT_METADATA,
    INIT_VAR_METADATA,
    FLATTEN_METADATA,
    POST_INIT_METADATA,
    PROPERTIES_METADATA,
    REQUIRED_METADATA,
    SKIP_METADATA,
}


@dataclass(frozen=True)
class Tagged(Generic[V]):
    metadata: Metadata = field(default_factory=MetadataImplem)
    alias: InitVar[Optional[str]] = None
    schema: InitVar[Optional[Schema]] = None
    deserialization: InitVar[Optional[ConvOrFunc]] = None
    serialization: InitVar[Optional[ConvOrFunc]] = None

    def __post_init__(
        self,
        alias: Optional[str],
        schema: Optional[Schema],
        deserialization: Optional[ConvOrFunc],
        serialization: Optional[ConvOrFunc],
    ):
        if self.metadata is None or isinstance(self.metadata, str):
            raise TypeError(
                "Tagged alias parameter is deprecated, use metadata instead"
            )
        if any(m is not None for m in (alias, schema, deserialization, serialization)):
            metadata = self.metadata
            warnings.warn(
                "Tagged keyword parameters are deprecated,"
                " use metadata parameter instead",
                DeprecationWarning,
            )
            if alias is not None:
                metadata |= alias_metadata(alias)
            if schema is not None:
                metadata |= schema
            if deserialization is not None or serialization is not None:
                metadata |= conversion(deserialization, serialization)
            object.__setattr__(self, "metadata", metadata)
        if self.metadata.keys() & INVALID_METADATA:
            raise TypeError("Invalid metadata in a TaggedUnion field")

    @overload
    def __get__(self, instance: None, owner: Type[T]) -> Tag[T, V]:
        ...

    @overload
    def __get__(self, instance: Any, owner) -> Union[V, UndefinedType]:
        ...

    def __get__(self, instance, owner):
        raise NotImplementedError


class TaggedUnion:
    def __init__(self, **kwargs):
        if len(kwargs) != 1:
            raise ValueError("TaggedUnion constructor expects only one field")
        tags = getattr(self, TAGS_ATTR)
        for tag in tags:
            setattr(self, tag, Undefined)
        for tag, value in kwargs.items():
            if tag not in tags:
                raise TypeError(f"{type(self)} has no tag {tag}")
            setattr(self, tag, value)

    def __repr__(self):
        tag, value = get_tagged(self)
        return f"{type(self).__name__}({tag}={value!r})"

    @wrap_generic_init_subclass
    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        tags = set(getattr(cls, TAGS_ATTR, ()))
        types = get_type_hints(cls, include_extras=True)
        for tag, tp in types.items():
            if get_origin2(tp) == Tagged:
                tagged = cls.__dict__.get(tag, Tagged())
                setattr(cls, tag, field(default=Undefined, metadata=tagged.metadata))
                cls.__annotations__[tag] = Union[
                    get_args2(types[tag])[0], UndefinedType
                ]
                tags.add(tag)
            elif tag not in tags:
                if get_origin2(tp) != ClassVar:
                    cls.__annotations__[tag] = ClassVar[tp]
                else:
                    raise TypeError(
                        "Only Tagged or ClassVar fields are allowed in TaggedUnion"
                    )
        setattr(cls, TAGS_ATTR, tags)
        schema(min_props=1, max_props=1)(dataclass(init=False, repr=False)(cls))
        for tag in tags:
            setattr(cls, tag, Tag(tag, cls))


def get_tagged(tagged_union: TaggedUnion) -> Tuple[str, Any]:
    defined = {
        tag: getattr(tagged_union, tag)
        for tag in getattr(tagged_union, TAGS_ATTR)
        if getattr(tagged_union, tag) is not Undefined
    }
    return next(iter(defined.items()))



from typing import Any, Callable, Dict, Type, TypeVar, Union

from apischema.json_schema.types import bad_type
from apischema.types import NoneType

T = TypeVar("T")

Coercer = Callable[[Type[T], Any], T]

_bool_pairs = (
    ("0", "1"),
    ("f", "t"),
    ("n", "y"),
    ("no", "yes"),
    ("false", "true"),
    ("off", "on"),
    ("ko", "ok"),
)
STR_TO_BOOL: Dict[str, bool] = {}
for false, true in _bool_pairs:
    for s, value in ((false, False), (true, True)):
        STR_TO_BOOL[s.lower()] = value
STR_NONE_VALUES = {""}


def coerce(cls: Type[T], data: Any) -> T:
    if cls is NoneType:
        if data is None or data in STR_NONE_VALUES:
            return None  # type: ignore
        else:
            raise bad_type(data, cls)
    elif isinstance(data, cls):
        return data
    elif cls is bool:
        if isinstance(data, str):
            return STR_TO_BOOL[data.lower()]  # type: ignore
        elif isinstance(data, int):
            return bool(data)  # type: ignore
        else:
            raise bad_type(data, cls)
    elif cls in (int, float):
        try:
            return cls(data)  # type: ignore
        except ValueError:
            raise bad_type(data, cls)
    elif cls is str:
        if isinstance(data, (int, float)) and not isinstance(data, bool):
            return str(data)  # type: ignore
        else:
            raise bad_type(data, cls)
    else:
        raise bad_type(data, cls)


Coerce = Union[bool, Coercer]



__all__ = [
    "alias",
    "conversion",
    "default_as_set",
    "fall_back_on_default",
    "flatten",
    "init_var",
    "none_as_undefined",
    "order",
    "post_init",
    "properties",
    "required",
    "schema",
    "skip",
    "validators",
]

import sys
import warnings

from apischema.aliases import alias
from apischema.ordering import order
from apischema.schemas import schema
from .implem import (
    conversion,
    default_as_set,
    fall_back_on_default,
    flatten,
    init_var,
    none_as_undefined,
    post_init,
    properties,
    required,
    skip,
    validators,
)

if sys.version_info >= (3, 7):

    def __getattr__(name):
        for deprecated in ("merged", "flattened"):
            if name == deprecated:
                warnings.warn(
                    f"apischema.metadata.{deprecated} is deprecated, "
                    "use apischema.metadata.flatten instead",
                    DeprecationWarning,
                )
                return flatten
            raise AttributeError(f"module {__name__} has no attribute {name}")


else:
    from .implem import flattened, merged  # noqa: F401



import collections.abc
import warnings
from contextlib import suppress
from dataclasses import dataclass
from typing import Any, Callable, MutableMapping, NamedTuple, Optional, TypeVar, Union

from apischema.cache import CacheAwareDict
from apischema.types import AnyType, PRIMITIVE_TYPES
from apischema.typing import get_args, get_origin, is_named_tuple, is_type_var
from apischema.utils import has_type_vars, merge_opts, replace_builtins


class TypeName(NamedTuple):
    json_schema: Optional[str] = None
    graphql: Optional[str] = None


NameOrFactory = Union[str, None, Callable[..., Optional[str]]]


def _apply_args(name_or_factory: NameOrFactory, *args) -> Optional[str]:
    return name_or_factory(*args) if callable(name_or_factory) else name_or_factory


_type_names: MutableMapping[AnyType, "TypeNameFactory"] = CacheAwareDict({})

T = TypeVar("T")


@dataclass(frozen=True)
class TypeNameFactory:
    json_schema: NameOrFactory
    graphql: NameOrFactory

    def __call__(self, tp: T) -> T:
        self.check_type(tp)
        _type_names[replace_builtins(tp)] = self
        return tp

    def check_type(self, tp: AnyType):
        if is_type_var(tp):
            raise TypeError("TypeVar cannot have a type_name")
        if has_type_vars(tp):
            if get_args(tp):
                raise TypeError("Generic alias cannot have a type_name")
            elif isinstance(self.json_schema, str) or isinstance(self.graphql, str):
                raise TypeError(
                    "Unspecialized generic type must used factory type_name"
                )

    def to_type_name(self, tp: AnyType, *args) -> TypeName:
        self.check_type(tp)
        return TypeName(
            _apply_args(self.json_schema, tp, *args),
            _apply_args(self.graphql, tp, *args),
        )


def type_name(
    ref: NameOrFactory = None,
    *,
    json_schema: NameOrFactory = None,
    graphql: NameOrFactory = None,
) -> TypeNameFactory:
    return TypeNameFactory(json_schema or ref, graphql or ref)


no_type_name = {*PRIMITIVE_TYPES, Any}


def default_type_name(tp: AnyType) -> Optional[TypeName]:
    if (
        hasattr(tp, "__name__")
        and not get_args(tp)
        and not has_type_vars(tp)
        and tp not in no_type_name
        and (
            not isinstance(tp, type)
            or not issubclass(tp, collections.abc.Collection)
            or is_named_tuple(tp)
        )
    ):
        return TypeName(tp.__name__, tp.__name__)
    else:
        return None


def get_type_name(tp: AnyType) -> TypeName:
    from apischema import settings

    tp = replace_builtins(tp)
    with suppress(KeyError, TypeError):
        return _type_names[tp].to_type_name(tp)
    origin, args = get_origin(tp), get_args(tp)
    if args and not has_type_vars(tp):
        with suppress(KeyError, TypeError):
            return _type_names[origin].to_type_name(origin, *args)
    return settings.default_type_name(tp) or TypeName()


@merge_opts
def merge_type_name(default: TypeName, override: TypeName) -> TypeName:
    return TypeName(
        override.json_schema or default.json_schema, override.graphql or default.graphql
    )


def schema_ref(ref: Optional[str]) -> Callable[[T], T]:
    warnings.warn("schema_ref is deprecated, use type_name instead", DeprecationWarning)
    return type_name(ref)



import collections.abc
import inspect
import re
import sys
import warnings
from contextlib import contextmanager, suppress
from dataclasses import dataclass, is_dataclass
from enum import Enum
from functools import wraps
from types import MappingProxyType
from typing import (
    AbstractSet,
    Any,
    Awaitable,
    Callable,
    Collection,
    Container,
    Dict,
    Generic,
    Hashable,
    Iterable,
    Iterator,
    List,
    Mapping,
    MutableMapping,
    NoReturn,
    Optional,
    Sequence,
    Set,
    Tuple,
    Type,
    TypeVar,
    Union,
    cast,
)

from apischema.types import (
    AnyType,
    COLLECTION_TYPES,
    MAPPING_TYPES,
    OrderedDict,
    PRIMITIVE_TYPES,
)
from apischema.typing import (
    _collect_type_vars,
    generic_mro,
    get_args,
    get_origin,
    get_type_hints,
    is_annotated,
    is_type_var,
    is_union,
    typing_origin,
)

try:
    from apischema.typing import Annotated
except ImportError:
    Annotated = ...  # type: ignore

PREFIX = "_apischema_"

T = TypeVar("T")
U = TypeVar("U")


def identity(x: T) -> T:
    return x


Lazy = Callable[[], T]


@dataclass(frozen=True)  # dataclass enable equality check
class LazyValue(Generic[T]):
    default: T

    def __call__(self) -> T:
        return self.default


if sys.version_info <= (3, 7):  # pragma: no cover
    is_dataclass_ = is_dataclass

    def is_dataclass(obj) -> bool:
        return is_dataclass_(obj) and getattr(obj, "__origin__", None) is None


def is_hashable(obj: Any) -> bool:
    return isinstance(obj, collections.abc.Hashable)


def opt_or(opt: Optional[T], default: U) -> Union[T, U]:
    return opt if opt is not None else default


def to_hashable(data: Union[None, int, float, str, bool, list, dict]) -> Hashable:
    if isinstance(data, list):
        return tuple(map(to_hashable, data))
    if isinstance(data, dict):
        return tuple(sorted((to_hashable(k), to_hashable(v)) for k, v in data.items()))
    return data  # type: ignore


SNAKE_CASE_REGEX = re.compile(r"_([a-z\d])")
CAMEL_CASE_REGEX = re.compile(r"[a-z\d]([A-Z])")


def to_camel_case(s: str) -> str:
    return SNAKE_CASE_REGEX.sub(lambda m: m.group(1).upper(), s)


def to_snake_case(s: str) -> str:
    return CAMEL_CASE_REGEX.sub(lambda m: "_" + m.group(1).lower(), s)


def to_pascal_case(s: str) -> str:
    camel = to_camel_case(s)
    return camel[0].upper() + camel[1:] if camel else camel


MakeDataclassField = Union[Tuple[str, AnyType], Tuple[str, AnyType, Any]]


def merge_opts(
    func: Callable[[T, T], T]
) -> Callable[[Optional[T], Optional[T]], Optional[T]]:
    def wrapper(opt1, opt2):
        if opt1 is None:
            return opt2
        if opt2 is None:
            return opt1
        return func(opt1, opt2)

    return wrapper


K = TypeVar("K")
V = TypeVar("V")


@merge_opts
def merge_opts_mapping(m1: Mapping[K, V], m2: Mapping[K, V]) -> Mapping[K, V]:
    return {**m1, **m2}


def has_type_vars(tp: AnyType) -> bool:
    return is_type_var(tp) or bool(getattr(tp, "__parameters__", ()))


TV = AnyType  # TypeVar is not supported as a type
# 10 should be enough for all builtin types
_type_vars = [TypeVar(f"T{i}") for i in range(10)]


def get_parameters(tp: AnyType) -> Iterable[TV]:
    if hasattr(tp, "__parameters__"):
        return tp.__parameters__
    elif hasattr(tp, "__orig_bases__"):
        return _collect_type_vars(tp.__orig_bases__)
    elif is_type_var(tp):
        return (tp,)
    else:
        return _type_vars


def substitute_type_vars(tp: AnyType, substitution: Mapping[TV, AnyType]) -> AnyType:
    if is_type_var(tp):
        try:
            return substitution[tp]
        except KeyError:
            return Union[tp.__constraints__] if tp.__constraints__ else Any
    elif getattr(tp, "__parameters__", ()):
        return tp[tuple(substitution.get(p, p) for p in tp.__parameters__)]
    else:
        return tp


Func = TypeVar("Func", bound=Callable)


def typed_wraps(wrapped: Func) -> Callable[[Callable], Func]:
    return cast(Func, wraps(wrapped))


def is_subclass(tp: AnyType, base: AnyType) -> bool:
    tp, base = get_origin_or_type(tp), get_origin_or_type(base)
    return tp == base or (
        isinstance(tp, type) and isinstance(base, type) and issubclass(tp, base)
    )


def _annotated(tp: AnyType) -> AnyType:
    return get_args(tp)[0] if is_annotated(tp) else tp


def get_origin_or_type(tp: AnyType) -> AnyType:
    origin = get_origin(tp)
    return origin if origin is not None else tp


def get_origin2(tp: AnyType) -> Optional[Type]:
    return get_origin(_annotated(tp))


def get_args2(tp: AnyType) -> Tuple[AnyType, ...]:
    return get_args(_annotated(tp))


def get_origin_or_type2(tp: AnyType) -> AnyType:
    tp2 = _annotated(tp)
    origin = get_origin(tp2)
    return origin if origin is not None else tp2


def keep_annotations(tp: AnyType, annotated: AnyType) -> AnyType:
    return Annotated[(tp, *get_args(annotated)[1:])] if is_annotated(annotated) else tp


def with_parameters(tp: AnyType) -> AnyType:
    return tp[tp.__parameters__] if getattr(tp, "__parameters__", ()) else tp


def is_union_of(tp: AnyType, of: AnyType) -> bool:
    return tp == of or (is_union(get_origin_or_type2(tp)) and of in get_args2(tp))


if sys.version_info < (3, 7):
    LIST_ORIGIN = List
    SET_ORIGIN = Set
    TUPLE_ORIGIN = Tuple
    DICT_ORIGIN = Dict
else:
    LIST_ORIGIN = typing_origin(list)
    SET_ORIGIN = typing_origin(set)
    TUPLE_ORIGIN = typing_origin(tuple)
    DICT_ORIGIN = typing_origin(dict)


def replace_builtins(tp: AnyType) -> AnyType:
    origin = get_origin2(tp)
    if origin is None:
        return tp
    args = tuple(map(replace_builtins, get_args2(tp)))
    replacement: Any
    if origin in COLLECTION_TYPES:
        if issubclass(origin, collections.abc.Set):
            replacement = SET_ORIGIN
        elif issubclass(origin, tuple) and (len(args) < 2 or args[1] is not ...):
            replacement = TUPLE_ORIGIN
        else:
            replacement = LIST_ORIGIN
    elif origin in MAPPING_TYPES:
        replacement = DICT_ORIGIN
    elif is_union(origin):
        replacement = Union
    else:
        replacement = typing_origin(origin)
    res = replacement[args] if args else replacement
    return keep_annotations(res, tp)


def sort_by_annotations_position(
    cls: Type, elts: Collection[T], key: Callable[[T], str]
) -> List[T]:
    annotations: Dict[str, Any] = OrderedDict()
    for base in reversed(cls.__mro__):
        annotations.update(getattr(base, "__annotations__", ()))
    positions = {key: i for i, key in enumerate(annotations)}
    return sorted(elts, key=lambda elt: positions.get(key(elt), len(positions)))


def stop_signature_abuse() -> NoReturn:
    raise TypeError("Stop signature abuse")


empty_dict: Mapping[str, Any] = MappingProxyType({})

ITERABLE_TYPES = {
    *COLLECTION_TYPES,
    *MAPPING_TYPES,
    Iterable,
    collections.abc.Iterable,
    Container,
    collections.abc.Container,
}


def subtyping_substitution(
    supertype: AnyType, subtype: AnyType
) -> Tuple[Mapping[AnyType, AnyType], Mapping[AnyType, AnyType]]:
    if not get_args(subtype) and not isinstance(subtype, type):
        return {}, {}
    supertype, subtype = with_parameters(supertype), with_parameters(subtype)
    supertype_to_subtype, subtype_to_supertype = {}, {}
    super_origin = get_origin_or_type2(supertype)
    for base in generic_mro(subtype):
        base_origin = get_origin_or_type2(base)
        if base_origin == super_origin or (
            base_origin in ITERABLE_TYPES and super_origin in ITERABLE_TYPES
        ):
            for base_arg, super_arg in zip(get_args2(base), get_args2(supertype)):
                if is_type_var(super_arg):
                    supertype_to_subtype[super_arg] = base_arg
                if is_type_var(base_arg):
                    subtype_to_supertype[base_arg] = super_arg
            break
    return supertype_to_subtype, subtype_to_supertype


def literal_values(values: Sequence[Any]) -> Sequence[Any]:
    primitive_values = [v.value if isinstance(v, Enum) else v for v in values]
    if any(not isinstance(v, PRIMITIVE_TYPES) for v in primitive_values):
        raise TypeError("Only primitive types are supported for Literal/Enum")
    return primitive_values


awaitable_origin = get_origin(Awaitable[Any])


def is_async(func: Callable, types: Mapping[str, AnyType] = None) -> bool:
    wrapped_func = func
    while hasattr(wrapped_func, "__wrapped__"):
        wrapped_func = wrapped_func.__wrapped__  # type: ignore
    if inspect.iscoroutinefunction(wrapped_func):
        return True
    if types is None:
        try:
            types = get_type_hints(func)
        except Exception:
            types = {}
    return get_origin_or_type2(types.get("return")) == awaitable_origin


@contextmanager
def context_setter(obj: Any):
    dict_copy = obj.__dict__.copy()
    try:
        yield
    finally:
        obj.__dict__.clear()
        obj.__dict__.update(dict_copy)


def wrap_generic_init_subclass(init_subclass: Func) -> Func:
    if sys.version_info >= (3, 7):
        return init_subclass

    @wraps(init_subclass)
    def wrapper(cls, **kwargs):
        if getattr(cls, "__origin__", None) is not None:
            super(cls).__init_subclass__(**kwargs)
            return
        init_subclass(cls, **kwargs)

    return wrapper


# # Because hash of generic classes is changed by metaclass after __init_subclass__
# # classes registered in global dictionaries are no more accessible. Here is a dictionary
# # wrapper to fix this issue
if sys.version_info < (3, 7):
    K = TypeVar("K")
    V = TypeVar("V")

    class KeyWrapper:
        def __init__(self, key):
            self.key = key

        def __eq__(self, other):
            return self.key == self.key

        def __hash__(self):
            return hash(
                id(self.key)
                if getattr(self.key, "__origin__", ...) is None
                else self.key
            )

    class type_dict_wrapper(MutableMapping[K, V]):
        def __init__(self, wrapped: Dict[K, V]):
            self.wrapped = cast(Dict[KeyWrapper, V], wrapped)

        def __delitem__(self, key: K) -> None:
            del self.wrapped[KeyWrapper(key)]

        def __getitem__(self, key: K) -> V:
            return self.wrapped[KeyWrapper(key)]

        def __iter__(self) -> Iterator[K]:
            return iter(wrapper.key for wrapper in list(self.wrapped))

        def __len__(self) -> int:
            return len(self.wrapped)

        def __setitem__(self, key: K, value: V):
            self.wrapped[KeyWrapper(key)] = value


else:
    M = TypeVar("M", bound=MutableMapping)

    def type_dict_wrapper(wrapped: M) -> M:
        return wrapped


def deprecate_kwargs(
    parameters_map: Mapping[str, Optional[str]]
) -> Callable[[Func], Func]:
    def decorator(func: Func) -> Func:
        wrapped = func.__init__ if isinstance(func, type) else func  # type: ignore

        def wrapper(*args, **kwargs):
            for param, replacement in parameters_map.items():
                if param in kwargs:
                    instead = f", use '{replacement}' instead" if replacement else ""
                    warnings.warn(
                        f"{func.__name__} parameter '{param}' is deprecated{instead}",
                        DeprecationWarning,
                    )
                    arg = kwargs.pop(param)
                    if replacement:
                        kwargs[replacement] = kwargs.get(replacement, arg)
            return wrapped(*args, **kwargs)

        if isinstance(func, type):
            func.__init__ = wraps(func.__init__)(wrapper)  # type: ignore
            return cast(Func, func)
        else:
            return cast(Func, wraps(func)(wrapper))

    return decorator


def as_predicate(
    collection_or_predicate: Union[Collection[T], Callable[[T], bool]]
) -> Callable[[T], bool]:
    if not isinstance(collection_or_predicate, Collection):
        return collection_or_predicate
    collection = collection_or_predicate
    if not isinstance(collection, AbstractSet):
        with suppress(Exception):
            collection = set(collection)

    def wrapper(elt: T) -> bool:
        try:
            return elt in collection
        except Exception:
            return False

    return wrapper



import warnings
from dataclasses import (  # type: ignore
    Field,
    InitVar,
    _FIELDS,
    _FIELD_CLASSVAR,
    make_dataclass,
)
from enum import Enum
from types import MappingProxyType
from typing import (
    Any,
    Collection,
    Generic,
    Mapping,
    Sequence,
    Tuple,
    Type,
    TypeVar,
    Union,
)

from apischema.types import (
    AnyType,
    COLLECTION_TYPES,
    MAPPING_TYPES,
    OrderedDict,
    PRIMITIVE_TYPES,
)
from apischema.typing import (
    get_args,
    get_origin,
    get_type_hints,
    get_type_hints2,
    is_annotated,
    is_literal,
    is_named_tuple,
    is_type_var,
    is_typed_dict,
    is_union,
    required_keys,
)
from apischema.utils import PREFIX, get_origin_or_type, has_type_vars, is_dataclass

try:
    from apischema.typing import Annotated
except ImportError:
    Annotated = ...  # type: ignore

TUPLE_TYPE = get_origin(Tuple[Any])


def type_hints_cache(obj) -> Mapping[str, AnyType]:
    return get_type_hints2(obj)


def dataclass_types_and_fields(
    tp: AnyType,
) -> Tuple[Mapping[str, AnyType], Sequence[Field], Sequence[Field]]:
    from apischema.metadata.keys import INIT_VAR_METADATA

    cls = get_origin_or_type(tp)
    assert is_dataclass(cls)
    types = get_type_hints2(tp)
    fields, init_fields = [], []
    for field in getattr(cls, _FIELDS).values():
        assert isinstance(field, Field)
        if field._field_type == _FIELD_CLASSVAR:  # type: ignore
            continue
        field_type = types[field.name]
        if isinstance(field_type, InitVar):
            types[field.name] = field_type.type  # type: ignore
            init_fields.append(field)
        elif field_type is InitVar:
            metadata = getattr(cls, _FIELDS)[field.name].metadata
            if INIT_VAR_METADATA not in metadata:
                raise TypeError("Before 3.8, InitVar requires init_var metadata")
            init_field = (PREFIX, metadata[INIT_VAR_METADATA], ...)
            tmp_cls = make_dataclass("Tmp", [init_field], bases=(cls,))  # type: ignore
            types[field.name] = get_type_hints(tmp_cls, include_extras=True)[PREFIX]
            if has_type_vars(types[field.name]):
                raise TypeError("Generic InitVar are not supported before 3.8")
            init_fields.append(field)
        else:
            fields.append(field)
    # Use immutable return because of cache
    return MappingProxyType(types), tuple(fields), tuple(init_fields)


class Unsupported(TypeError):
    def __init__(self, tp: AnyType):
        self.type = tp

    @property
    def cls(self) -> AnyType:
        warnings.warn(
            "Unsupported.cls is deprecated, use Unsupported.type instead",
            DeprecationWarning,
        )
        return self.type


Result = TypeVar("Result", covariant=True)


class Visitor(Generic[Result]):
    def annotated(self, tp: AnyType, annotations: Sequence[Any]) -> Result:
        if Unsupported in annotations:
            raise Unsupported(Annotated[(tp, *annotations)])  # type: ignore
        return self.visit(tp)

    def any(self) -> Result:
        raise NotImplementedError

    def collection(self, cls: Type[Collection], value_type: AnyType) -> Result:
        raise NotImplementedError

    def dataclass(
        self,
        tp: AnyType,
        types: Mapping[str, AnyType],
        fields: Sequence[Field],
        init_vars: Sequence[Field],
    ) -> Result:
        raise NotImplementedError

    def enum(self, cls: Type[Enum]) -> Result:
        raise NotImplementedError

    def literal(self, values: Sequence[Any]) -> Result:
        raise NotImplementedError

    def mapping(
        self, cls: Type[Mapping], key_type: AnyType, value_type: AnyType
    ) -> Result:
        raise NotImplementedError

    def named_tuple(
        self, tp: AnyType, types: Mapping[str, AnyType], defaults: Mapping[str, Any]
    ) -> Result:
        raise NotImplementedError

    def new_type(self, tp: AnyType, super_type: AnyType) -> Result:
        return self.visit(super_type)

    def primitive(self, cls: Type) -> Result:
        raise NotImplementedError

    def subprimitive(self, cls: Type, superclass: Type) -> Result:
        return self.primitive(superclass)

    def tuple(self, types: Sequence[AnyType]) -> Result:
        raise NotImplementedError

    def typed_dict(
        self, tp: AnyType, types: Mapping[str, AnyType], required_keys: Collection[str]
    ) -> Result:
        raise NotImplementedError

    def union(self, alternatives: Sequence[AnyType]) -> Result:
        raise NotImplementedError

    def unsupported(self, tp: AnyType) -> Result:
        raise Unsupported(tp)

    def visit(self, tp: AnyType) -> Result:
        origin, args = get_origin_or_type(tp), get_args(tp)
        if args:
            if is_annotated(tp):
                return self.annotated(args[0], args[1:])
            if is_union(origin):
                return self.union(args[0]) if len(args) == 1 else self.union(args)
            if origin is TUPLE_TYPE:
                if len(args) < 2 or args[1] is not ...:
                    return self.tuple(args)
            if origin in COLLECTION_TYPES:
                return self.collection(origin, args[0])
            if origin in MAPPING_TYPES:
                return self.mapping(origin, args[0], args[1])
            if is_literal(tp):  # pragma: no cover py37+
                return self.literal(args)
        if origin in PRIMITIVE_TYPES:
            return self.primitive(origin)
        if is_dataclass(origin):
            return self.dataclass(tp, *dataclass_types_and_fields(tp))  # type: ignore
        if hasattr(origin, "__supertype__"):
            return self.new_type(origin, origin.__supertype__)
        if origin is Any:
            return self.any()
        if origin in COLLECTION_TYPES:
            return self.collection(origin, Any)
        if origin in MAPPING_TYPES:
            return self.mapping(origin, Any, Any)
        if isinstance(origin, type):
            if issubclass(origin, Enum):
                return self.enum(origin)
            for primitive in PRIMITIVE_TYPES:
                if issubclass(origin, primitive):
                    return self.subprimitive(origin, primitive)
            # NamedTuple
            if is_named_tuple(origin):
                if hasattr(origin, "__annotations__"):
                    types = type_hints_cache(origin)
                elif hasattr(origin, "__field_types"):  # pragma: no cover
                    types = origin.__field_types  # type: ignore
                else:  # pragma: no cover
                    types = OrderedDict((f, Any) for f in origin._fields)  # type: ignore  # noqa: E501
                return self.named_tuple(
                    origin, types, origin._field_defaults  # type: ignore
                )
        if is_literal(origin):  # pragma: no cover py36
            return self.literal(origin.__values__)  # type: ignore
        if is_typed_dict(origin):
            return self.typed_dict(
                origin, type_hints_cache(origin), required_keys(origin)
            )
        if is_type_var(origin):
            if origin.__constraints__:
                return self.visit(Union[origin.__constraints__])
            else:
                return self.any()
        return self.unsupported(tp)



from dataclasses import dataclass
from functools import partial
from types import FunctionType, MethodType
from typing import Any, Mapping, Optional, TYPE_CHECKING, Type, TypeVar

from apischema.fields import FIELDS_SET_ATTR
from apischema.objects import object_fields

if TYPE_CHECKING:
    from apischema.validation.validators import Validator

MOCK_FIELDS_FIELD = "__mock_fields__"
MOCK_CLS_FIELD = "__mock_cls__"


class NonTrivialDependency(Exception):
    def __init__(self, attr: str):
        self.attr = attr
        self.validator: Optional["Validator"] = None


@dataclass(init=False)
class ValidatorMock:
    def __init__(self, cls: Type, values: Mapping[str, Any]):
        self.cls = cls
        self.values = values

    def __getattribute__(self, name: str) -> Any:
        values = super().__getattribute__("values")
        if name in values:
            return values[name]
        cls = super().__getattribute__("cls")
        fields = object_fields(cls, deserialization=True)
        if name in fields:
            if fields[name].required:
                raise NonTrivialDependency(name)
            return fields[name].get_default()
        if name == "__class__":
            return cls
        if name == "__dict__":
            return {**values, FIELDS_SET_ATTR: set(values)}
        if name == FIELDS_SET_ATTR:
            return set(values)
        if hasattr(cls, name):
            member = getattr(cls, name)
            # for classmethod (staticmethod are not handled)
            if isinstance(member, MethodType):
                return member
            if isinstance(member, FunctionType):
                return partial(member, self)
            if isinstance(member, property):
                return member.fget(self)  # type: ignore
            return member
        raise NonTrivialDependency(name)


T = TypeVar("T")



from dataclasses import Field, MISSING
from typing import Any, Collection, Mapping, Optional, Sequence

from apischema.aliases import Aliaser, get_class_aliaser
from apischema.conversions.conversions import AnyConversion
from apischema.dataclasses import replace
from apischema.metadata.keys import ALIAS_METADATA
from apischema.objects.fields import FieldKind, MISSING_DEFAULT, ObjectField
from apischema.types import AnyType, Undefined
from apischema.typing import get_args
from apischema.utils import get_origin_or_type, get_parameters, substitute_type_vars
from apischema.visitor import Result, Visitor


def object_field_from_field(
    field: Field, field_type: AnyType, init_var: bool
) -> ObjectField:
    required = field.default is MISSING and field.default_factory is MISSING  # type: ignore
    if init_var:
        kind = FieldKind.WRITE_ONLY
    elif not field.init:
        kind = FieldKind.READ_ONLY
    else:
        kind = FieldKind.NORMAL
    return ObjectField(
        field.name,
        field_type,
        required,
        field.metadata,
        default=field.default,
        default_factory=field.default_factory,  # type: ignore
        kind=kind,
    )


def _override_alias(field: ObjectField, aliaser: Aliaser) -> ObjectField:
    if field.override_alias:
        return replace(
            field,
            metadata={**field.metadata, ALIAS_METADATA: aliaser(field.alias)},
            default=MISSING_DEFAULT,
        )
    else:
        return field


class ObjectVisitor(Visitor[Result]):
    _field_kind_filtered: Optional[FieldKind] = None

    def _field_conversion(self, field: ObjectField) -> Optional[AnyConversion]:
        raise NotImplementedError

    def _skip_field(self, field: ObjectField) -> bool:
        raise NotImplementedError

    @staticmethod
    def _default_fields(cls: type) -> Optional[Sequence[ObjectField]]:
        from apischema import settings

        return settings.default_object_fields(cls)

    def _override_fields(
        self, tp: AnyType, fields: Sequence[ObjectField]
    ) -> Sequence[ObjectField]:

        origin = get_origin_or_type(tp)
        if isinstance(origin, type):
            default_fields = self._default_fields(origin)
            if default_fields is not None:
                if get_args(tp):
                    sub = dict(zip(get_parameters(origin), get_args(tp)))
                    default_fields = [
                        replace(f, type=substitute_type_vars(f.type, sub))
                        for f in default_fields
                    ]
                return default_fields
        return fields

    def _object(self, tp: AnyType, fields: Sequence[ObjectField]) -> Result:
        fields = [f for f in fields if not self._skip_field(f)]
        aliaser = get_class_aliaser(get_origin_or_type(tp))
        if aliaser is not None:
            fields = [_override_alias(f, aliaser) for f in fields]
        return self.object(tp, fields)

    def dataclass(
        self,
        tp: AnyType,
        types: Mapping[str, AnyType],
        fields: Sequence[Field],
        init_vars: Sequence[Field],
    ) -> Result:
        by_name = {
            f.name: object_field_from_field(f, types[f.name], init_var)
            for field_group, init_var in [(fields, False), (init_vars, True)]
            for f in field_group
        }
        object_fields = [
            by_name[name]
            for name in types
            if name in by_name and by_name[name].kind != self._field_kind_filtered
        ]
        return self._object(tp, self._override_fields(tp, object_fields))

    def object(self, tp: AnyType, fields: Sequence[ObjectField]) -> Result:
        raise NotImplementedError

    def named_tuple(
        self, tp: AnyType, types: Mapping[str, AnyType], defaults: Mapping[str, Any]
    ) -> Result:
        fields = [
            ObjectField(name, type_, name not in defaults, default=defaults.get(name))
            for name, type_ in types.items()
        ]
        return self._object(tp, self._override_fields(tp, fields))

    def typed_dict(
        self, tp: AnyType, types: Mapping[str, AnyType], required_keys: Collection[str]
    ) -> Result:
        fields = [
            ObjectField(name, type_, name in required_keys, default=Undefined)
            for name, type_ in types.items()
        ]
        return self._object(tp, self._override_fields(tp, fields))

    def unsupported(self, tp: AnyType) -> Result:
        dummy: list = []
        fields = self._override_fields(tp, dummy)
        return super().unsupported(tp) if fields is dummy else self._object(tp, fields)


class DeserializationObjectVisitor(ObjectVisitor[Result]):
    _field_kind_filtered = FieldKind.READ_ONLY

    @staticmethod
    def _field_conversion(field: ObjectField) -> Optional[AnyConversion]:
        return field.deserialization

    @staticmethod
    def _skip_field(field: ObjectField) -> bool:
        return field.skip.deserialization


class SerializationObjectVisitor(ObjectVisitor[Result]):
    _field_kind_filtered = FieldKind.WRITE_ONLY

    @staticmethod
    def _field_conversion(field: ObjectField) -> Optional[AnyConversion]:
        return field.serialization

    @staticmethod
    def _skip_field(field: ObjectField) -> bool:
        return field.skip.serialization



from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from typing import (
    Awaitable,
    ClassVar,
    Collection,
    Dict,
    Generic,
    List,
    Type,
    TypeVar,
    Union,
    cast,
)

import graphql

from apischema import deserialize, deserializer, serialize, serializer, type_name
from apischema.graphql import ID, interface, resolver
from apischema.metadata import skip
from apischema.ordering import order
from apischema.type_names import get_type_name
from apischema.typing import generic_mro, get_args, get_origin
from apischema.utils import PREFIX, has_type_vars, wrap_generic_init_subclass

ID_TYPE_ATTR = f"{PREFIX}id_type"


class InvalidGlobalId(Exception):
    def __init__(self, value: str):
        self.value = value

    def __str__(self):
        return f"{self.value} is not a valid global id"


class NotANode(Exception):
    def __init__(self, node_type: str):
        self.node_type = node_type

    def __str__(self):
        return f"{self.node_type} is not a Node"


Node_ = TypeVar("Node_", bound="Node")


@dataclass
class GlobalId(Generic[Node_]):
    id: str
    node_type: Type[Node_]


@deserializer
def deserialize_global_id(global_id: ID) -> GlobalId:
    try:
        node_key, id = global_id.split(":")
    except ValueError:
        raise InvalidGlobalId(global_id) from None
    try:
        return GlobalId(id, _nodes[node_key])
    except KeyError:
        raise NotANode(node_key) from None


@serializer
def serialize_global_id(global_id: GlobalId) -> ID:
    return ID(f"{global_id.node_type._node_key()}:{global_id.id}")


Id = TypeVar("Id")


@type_name(graphql=lambda *_: "Node")
@interface
@dataclass  # type: ignore
class Node(Generic[Id], ABC):
    id: Id = field(metadata=skip)
    global_id: ClassVar[property]

    @property  # type: ignore
    def global_id(self: Node_) -> GlobalId[Node_]:
        return self.id_to_global(self.id)

    @classmethod
    def id_from_global(cls: Type[Node_], global_id: GlobalId[Node_]) -> Id:
        if global_id.node_type != cls:
            raise ValueError(
                f"Expected {cls.__name__} global id,"
                f" found {global_id.node_type.__name__} global id"
            )
        id_type = getattr(cls, ID_TYPE_ATTR)
        # Use coercion to handle integer id
        return cast(Id, deserialize(id_type, global_id.id, coerce=True))

    @classmethod
    def id_to_global(cls: Type[Node_], id: Id) -> GlobalId[Node_]:
        return GlobalId(str(serialize(getattr(cls, ID_TYPE_ATTR), id)), cls)

    @classmethod
    @abstractmethod
    def get_by_id(
        cls: Type[Node_], id: Id, info: graphql.GraphQLResolveInfo = None
    ) -> Union[Node_, Awaitable[Node_]]:
        raise NotImplementedError

    @classmethod
    def _node_key(cls) -> str:
        node_name = get_type_name(cls).graphql
        if node_name is None:
            raise TypeError(f"Node {cls} has no type_name registered")
        return node_name

    @wrap_generic_init_subclass
    def __init_subclass__(cls, not_a_node: bool = False, **kwargs):
        super().__init_subclass__(**kwargs)  # type: ignore
        if not not_a_node:
            _tmp_nodes.append(cls)


resolver("id", order=order(-1))(
    Node.global_id
)  # cannot directly decorate property because py36

_tmp_nodes: List[Type[Node]] = []
_nodes: Dict[str, Type[Node]] = {}


def process_node(node_cls: Type[Node]):
    if has_type_vars(node_cls) or node_cls.get_by_id is Node.get_by_id:
        return
    for base in node_cls.__mro__:
        if base != Node and Node.get_by_id.__name__ in base.__dict__:
            if not isinstance(
                base.__dict__[Node.get_by_id.__name__], (classmethod, staticmethod)
            ):
                raise TypeError(
                    f"{node_cls.__name__}.get_by_id must be a"
                    f" classmethod/staticmethod"
                )
            break
    for base in generic_mro(node_cls):
        if get_origin(base) == Node:
            setattr(node_cls, ID_TYPE_ATTR, get_args(base)[0])
            _nodes[node_cls._node_key()] = node_cls
            break
    else:
        raise TypeError("Node type parameter Id must be specialized")


def nodes() -> Collection[Type[Node]]:
    for node_cls in _tmp_nodes:
        process_node(node_cls)
    return list(_nodes.values())


def node(id: ID, info: graphql.GraphQLResolveInfo = None) -> Node:
    global_id = deserialize_global_id(id)
    node_type = global_id.node_type
    return node_type.get_by_id(node_type.id_from_global(global_id), info)



import collections.abc
import sys
from enum import Enum, auto
from itertools import chain
from types import MappingProxyType
from typing import (
    AbstractSet,
    Any,
    Collection,
    Dict,
    FrozenSet,
    List,
    Mapping,
    MutableMapping,
    MutableSequence,
    MutableSet,
    Sequence,
    Set,
    TYPE_CHECKING,
    Tuple,
    Type,
    Union,
)

AnyType = Any
NoneType: Type[None] = type(None)
Number = Union[int, float]

PRIMITIVE_TYPES = (str, int, bool, float, NoneType)
COLLECTION_TYPES = (
    Collection,
    collections.abc.Collection,
    Sequence,
    collections.abc.Sequence,
    Tuple,
    tuple,
    MutableSequence,
    collections.abc.MutableSequence,
    List,
    list,
    AbstractSet,
    collections.abc.Set,
    FrozenSet,
    frozenset,
    MutableSet,
    collections.abc.MutableSet,
    Set,
    set,
)
MAPPING_TYPES = (
    Mapping,
    collections.abc.Mapping,
    MutableMapping,
    collections.abc.MutableMapping,
    Dict,
    dict,
    MappingProxyType,
)


if sys.version_info >= (3, 7):  # pragma: no cover
    OrderedDict = dict
    ChainMap = collections.ChainMap
else:  # pragma: no cover
    OrderedDict = collections.OrderedDict

    class ChainMap(collections.ChainMap):
        def __iter__(self):
            return iter({k: None for k in chain.from_iterable(reversed(self.maps))})


class Metadata(Mapping[str, Any]):
    def __or__(self, other: Mapping[str, Any]) -> "Metadata":
        return MetadataImplem({**self, **other})

    def __ror__(self, other: Mapping[str, Any]) -> "Metadata":
        return MetadataImplem({**other, **self})


class MetadataMixin(Metadata):
    key: str

    def __getitem__(self, key):
        if key != self.key:
            raise KeyError(key)
        return self

    def __iter__(self):
        return iter((self.key,))

    def __len__(self):
        return 1


class MetadataImplem(dict, Metadata):  # type: ignore
    def __hash__(self):
        return hash(tuple(sorted(self.items())))


# Singleton type, see https://www.python.org/dev/peps/pep-0484/#id30
if TYPE_CHECKING:

    class UndefinedType(Enum):
        Undefined = auto()

    Undefined = UndefinedType.Undefined
else:

    class UndefinedType:
        def __new__(cls):
            return Undefined

        def __repr__(self):
            return "Undefined"

        def __str__(self):
            return "Undefined"

        def __bool__(self):
            return False

    Undefined = object.__new__(UndefinedType)



__all__ = [
    "ClientMutationId",
    "Connection",
    "Edge",
    "GlobalId",
    "Mutation",
    "Node",
    "PageInfo",
    "base64_encoding",
    "mutations",
    "node",
    "nodes",
]
from .connections import Connection, Edge, PageInfo
from .global_identification import GlobalId, Node, node, nodes
from .mutations import ClientMutationId, Mutation, mutations
from .utils import base64_encoding



__all__ = [
    "JsonSchemaVersion",
    "definitions_schema",
    "deserialization_schema",
    "serialization_schema",
]

from apischema.json_schema.schema import (
    definitions_schema,
    deserialization_schema,
    serialization_schema,
)
from .versions import JsonSchemaVersion



from collections import defaultdict
from dataclasses import dataclass
from functools import wraps
from inspect import Parameter, isclass, signature
from typing import (
    Any,
    Callable,
    Collection,
    Dict,
    Mapping,
    MutableMapping,
    NoReturn,
    Optional,
    Tuple,
    Type,
    TypeVar,
    Union,
    overload,
)

from apischema.cache import CacheAwareDict
from apischema.conversions.conversions import AnyConversion
from apischema.methods import method_registerer
from apischema.ordering import Ordering
from apischema.schemas import Schema
from apischema.types import AnyType, Undefined, UndefinedType
from apischema.typing import generic_mro, get_type_hints
from apischema.utils import (
    deprecate_kwargs,
    get_args2,
    get_origin_or_type,
    get_origin_or_type2,
    substitute_type_vars,
    subtyping_substitution,
)


@dataclass(frozen=True)
class SerializedMethod:
    func: Callable
    alias: str
    conversion: Optional[AnyConversion]
    error_handler: Optional[Callable]
    ordering: Optional[Ordering]
    schema: Optional[Schema]

    def error_type(self) -> AnyType:
        assert self.error_handler is not None
        types = get_type_hints(self.error_handler, include_extras=True)
        if "return" not in types:
            raise TypeError("Error handler must be typed")
        return types["return"]

    def return_type(self, return_type: AnyType) -> AnyType:
        if self.error_handler is not None:
            error_type = self.error_type()
            if error_type is not NoReturn:
                return Union[return_type, error_type]
        return return_type

    def types(self, owner: AnyType = None) -> Mapping[str, AnyType]:
        types = get_type_hints(self.func, include_extras=True)
        if "return" not in types:
            if isclass(self.func):
                types["return"] = self.func
            else:
                raise TypeError("Function must be typed")
        types["return"] = self.return_type(types["return"])
        if get_args2(owner):
            first_param = next(iter(signature(self.func).parameters))
            substitution, _ = subtyping_substitution(
                types.get(first_param, get_origin_or_type2(owner)), owner
            )
            types = {
                name: substitute_type_vars(tp, substitution)
                for name, tp in types.items()
            }
        return types


_serialized_methods: MutableMapping[Type, Dict[str, SerializedMethod]] = CacheAwareDict(
    defaultdict(dict)
)

S = TypeVar("S", bound=SerializedMethod)


def _get_methods(
    tp: AnyType, all_methods: Mapping[Type, Mapping[str, S]]
) -> Collection[Tuple[S, Mapping[str, AnyType]]]:
    result = {}
    for base in reversed(generic_mro(tp)):
        for name, method in all_methods[get_origin_or_type(base)].items():
            result[name] = (method, method.types(base))
    return result.values()


def get_serialized_methods(
    tp: AnyType,
) -> Collection[Tuple[SerializedMethod, Mapping[str, AnyType]]]:
    return _get_methods(tp, _serialized_methods)


ErrorHandler = Union[Callable, None, UndefinedType]


def none_error_handler(error: Exception, obj: Any, alias: str) -> None:
    return None


MethodOrProp = TypeVar("MethodOrProp", Callable, property)


@overload
def serialized(__method_or_property: MethodOrProp) -> MethodOrProp:
    ...


@overload
def serialized(
    alias: str = None,
    *,
    conversion: AnyConversion = None,
    error_handler: ErrorHandler = Undefined,
    order: Optional[Ordering] = None,
    schema: Schema = None,
    owner: Type = None,
) -> Callable[[MethodOrProp], MethodOrProp]:
    ...


@deprecate_kwargs({"conversions": "conversion"})
def serialized(
    __arg=None,
    *,
    alias: str = None,
    conversion: AnyConversion = None,
    error_handler: ErrorHandler = Undefined,
    order: Optional[Ordering] = None,
    schema: Schema = None,
    owner: Type = None,
):
    def register(func: Callable, owner: Type, alias2: str):
        alias2 = alias or alias2
        parameters = list(signature(func).parameters.values())
        for param in parameters[1:]:
            if (
                param.kind not in {Parameter.VAR_POSITIONAL, Parameter.VAR_KEYWORD}
                and param.default is Parameter.empty
            ):
                raise TypeError("Serialized method cannot have required parameter")
        error_handler2 = error_handler
        if error_handler is None:
            error_handler2 = none_error_handler
        if error_handler2 is Undefined:
            error_handler2 = None
        else:
            wrapped = func

            @wraps(wrapped)
            def func(self):
                try:
                    return wrapped(self)
                except Exception as error:
                    return error_handler(error, self, alias2)

        assert not isinstance(error_handler2, UndefinedType)
        _serialized_methods[owner][alias2] = SerializedMethod(
            func, alias2, conversion, error_handler2, order, schema
        )

    if isinstance(__arg, str):
        alias = __arg
        __arg = None
    return method_registerer(__arg, owner, register)
